###################################################################################################
#  KaptsjaGenerator creates Dynamically Captcha puzzles. They cen be served dynamically or as a pre-generated CaptCha set
#  Kaptsja is the Dutch phonetic pronunciation of the English word Captcha.
#
#  WARNING: BEFORE CHANGING ANYTHING IN THIS FILE MAKE A BACKUP SO YOU CAN RESTORE !
#  Read the comments in this file as they contain valuable information to understand the effects 
###################################################################################################

import sys
# sys.path.append("/path/to/your/package_or_module")
sys.path.append(r".")
# import pytesseract  # only used to parse generated Kaptsja Image false for analysis purposes only. Program tesseract must be installed!
from PIL import Image, ImageDraw, ImageFont  #,  ImageFilter
from bs4 import BeautifulSoup as bs  # prettify/complete broken HTML output
import random
import re
import string
from base64 import b64decode
import os
import glob
import time
import logging
import threading
from KaptsjaEncDec import Cripto
# import the configuration
from KaptsjaConfiguration import *
#
############################################################################################################################################################
#  SETTINGS  (start) See ./scripts/Kaptsja_configuration.py 
############################################################################################################################################################
# end of imports
logging.basicConfig(filename=log_file, level=logging.ERROR, \
    format='%(asctime)s %(levelname)-8s %(message)s', \
    datefmt='%Y-%m-%d %H:%M:%S')
# make directory for encryption key when it doesn't exists
if not os.path.exists(key_dir):
    os.makedirs(key_dir)
# make directory for temporary files when it doesn't exists
if not os.path.exists(work_dir):
    os.makedirs(work_dir)
##########################################
# START of validation of the configuration
##########################################  
syserrmsg = "Error! Look in log file."
# Function to validate 
# hexadecimal color code . 
def isValidHexaCode(str):
    # Regex to check valid 
    # hexadecimal color code.
    regex = "^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$"
    # Compile the ReGex 
    p = re.compile(regex)
    # If the string is empty 
    # return false
    if(str == None):
        return False
    # Return if the string 
    # matched the ReGex
    if(re.search(p, str)):
        return True
    else:
        return False
#
if type(active_captcha_model) != int:
    text = "active_captcha_model %s is not an integer. Change the value to 1, 2 or 3." % active_captcha_model
    logging.info(text)
    sys.exit(syserrmsg)
if not active_captcha_model in (1,2,3):
    active_captcha_model = 2  # default 2    
# create empty picture list
picture_list = []
allowed_file_extensions = (".png", ".jpg", ".jpeg", ".bmp", 'tiff', 'gif', 'eps', 'pcx')  # The dot is required! Use only lower case! See for more https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html
if randomlist == True: # all all
    for filename in glob.glob(randomlist_dir +'*'):
        infile_name, infile_ext = os.path.splitext(filename)
        if infile_ext.lower() in allowed_file_extensions:
            picture_list.append(filename)
else:
        picture_list.append(input_picture_file)
if not len(picture_list) > 0:
    text = "Are there pictures in %s?\nAre these settings defined/correct?\n input_picture, input_picture_file, randomlist, randomlist_dir." % randomlist_dir
    logging.error(text)
    sys.exit(syserrmsg) 
if os.path.exists(input_picture_file) == False:
    text = "input_picture_file %s does not exist in directory %s. Check media_dir setting and spelling of file name (upper/lower case sensitive) or put an own picture with the exact name in the directory." % (input_picture_file, media_dir)
    logging.error(text)
    sys.exit(syserrmsg)
if type(input_picture_rescale_width) != int:
    text = "Value input_picture_rescale_width %s must be an integer! Minimum 360 and maximum 598.\nValues > maximum are automatically reset to maximum." % input_picture_rescale_width
    logging.error(text)
    sys.exit(syserrmsg)
if 360 <= int(input_picture_rescale_width) <= 599 or int(input_picture_rescale_width) == 0: # when true we skip else
    pass
else:
    text = "Value input_picture_rescale_width %s should be minimum 360 and maximum 598.\nValues > maximum are automatically reset to maximum." % input_picture_rescale_width
    logging.error(text)
    sys.exit(syserrmsg)
if os.path.exists(font_textzone) == False:
    text = "Font file defined %s does not exist. Check font_textzone setting and spelling of file name (upper/lower case sensitive)." % font_textzone
    logging.error(text)
    sys.exit(syserrmsg)
if os.path.exists(font_circle) == False:
    text = "Font file defined %s does not exist. Check font_circle setting and spelling of file name (upper/lower case sensitive)." % font_circle
    logging.error(text)
    sys.exit(syserrmsg)
if type(font_pnts_circle) != int:
    text = "font_pnts_circle %s is not an integer. Change the value." % font_pnts_circle
    logging.error(text)
    sys.exit(syserrmsg)
if type(font_pnts_textzone) != int:
    text = "font_pnts_textzone %s is not an integer. Change the value." % font_pnts_textzone
    logging.error(text)
    sys.exit(syserrmsg)
if type(number_of_letters) != int:
    text = "number_of_letters %s is not an integer. Change the value." % number_of_letters
    logging.error(text)
    sys.exit(syserrmsg)
if type(number_of_digits) != int:
    text = "number_of_digits %s is not an integer. Change the value." % number_of_digits
    logging.error(text)
    sys.exit(syserrmsg)
if type(border_circle) != int:
    text = "border_circle %s is not an integer. Change the value." % border_circle
    logging.error(text)
    sys.exit(syserrmsg)
if type(border_line) != int:
    text = "border_line %s is not an integer. Change the value." % border_line
    logging.error(text)
    sys.exit(syserrmsg)
if type(max_retries) != int:
    text = "max_retries %s is not an integer. Change the value." % max_retries
    logging.error(text)
    sys.exit(syserrmsg)
if type(circle_color_trans) != int:
    text = "circle_color_trans %s is not an integer. Change the value." % circle_color_trans
    logging.error(text)
    sys.exit(syserrmsg)
if  circle_color_trans not in range(-1,101):
    text = "circle_color_trans %s must be in range 0 - 100. It reflects the percentage of transparency." % circle_color_trans
    logging.error(text)
    sys.exit(syserrmsg)
if type(char_circle_color_trans) != int:
    text = "char_circle_color_trans %s is not an integer. Change the value." % char_circle_color_trans
    logging.error(text)
    sys.exit(syserrmsg)
if  char_circle_color_trans not in range(-1,101):
    text = "char_circle_color_trans %s must be in range 0 - 100. It reflects the percentage of transparency." % char_circle_color_trans
    logging.error(text)
    sys.exit(syserrmsg)
if type(textzone_color_trans) != int:
    text = "textzone_color_trans %s is not an integer. Change the value." % textzone_color_trans
    logging.error(text)
    sys.exit(syserrmsg)
if  textzone_color_trans not in range(-1,101):
    text = "textzone_color_trans %s must be in range 0 - 100. It reflects the percentage of transparency." % textzone_color_trans
    logging.error(text)
    sys.exit(syserrmsg) 
if  not isValidHexaCode(str(textzone_text_color)):
    text = "textzone_text_color %s is not a correct hexadecimal color code" % textzone_text_color
    logging.error(text)
    sys.exit(syserrmsg) 
if  not isValidHexaCode(str(line_color)):
    text = "line_color %s is not a correct hexadecimal color code" % line_color
    logging.error(text)
    sys.exit(syserrmsg) 
if  not isValidHexaCode(str(textzone_bg_color)):
    text = "textzone_bg_color %s is not a correct hexadecimal color code" % textzone_bg_color
    logging.error(text)
    sys.exit(syserrmsg) 
if type(max_morphing_pie_slices) != int:
    text = "max_morphing_pie_slices %s is not an integer. Change the value." % max_morphing_pie_slices
    logging.error(text)
    sys.exit(syserrmsg)
if  max_morphing_pie_slices < 3:
    text = "max_morphing_pie_slices %s must be 3 or higher integer. Change the value." % max_morphing_pie_slices
    logging.error(text)
    sys.exit(syserrmsg)
if type(allowed_evals_per_char) != int:
    text = "allowed_evals_per_char %s is not an integer. Change the value." % allowed_evals_per_char
    logging.error(text)
    sys.exit(syserrmsg)
Unicode_Characters_found = False    
if  len(re.findall(r'[\u4e00-\u9fff]+', used_letters) ) > 0:
    Unicode_Characters_found = True
    if any([x in string.digits for x in used_letters]) == True:
        text = u"used_letters %s may not contain digits. Change the value." % used_letters.encode("utf8")
        logging.error(text)
        sys.exit(syserrmsg)
if  Unicode_Characters_found == False:
    if used_letters.isalpha() == False:
        text = u'used_letters %s may only contain alpha letters. Change the value!\nNote: Unicode characters like Chinese are also possible :-), but then the string in the configuration file must start with u"...." to identify it as Unicode. ' % ( used_letters.encode("utf8"))
        logging.error(text)
        sys.exit(syserrmsg)
if all([x in string.digits for x in used_digits]) == False:
    text = "used_digits %s may only contain digits. Change the value." % used_digits
    logging.error(text)
    sys.exit(syserrmsg)
if  number_of_letters + number_of_digits < 1:
    text = "number_of_letters + number_of_digits (%s and %s) must be minimal 1, else no Circles will be drawn. Increase the number of letters or digits or both." % (number_of_letters, number_of_digits)
    logging.error(text)
    sys.exit(syserrmsg)
if  number_of_letters + number_of_digits > 10:
    text = "number_of_letters + number_of_digits (%s and %s) is maximized 10 to avoid overcrowded Kaptsja's. Decrease the number of letters or digits or both." % (number_of_letters, number_of_digits)
    logging.error(text)
    sys.exit(syserrmsg)
if os.path.exists(gen_home_page_file) != True:
    text = "File %s does not exist in directory %s.\nLooking for %s.\nCheck html_dir setting and spelling of file name (upper/lower case sensitive) or create an own failure page with the exact name. Run KaptsjaHTMLpages.py." % (gen_home_page, html_dir, gen_home_page_file)
    logging.error(text)
    sys.exit(syserrmsg)
if os.path.exists(gen_failure_page_file) != True:
    text = "File %s does not exist in directory %s.\nLooking for %s.\nCheck html_dir setting and spelling of file name (upper/lower case sensitive) or create an own failure page with the exact name. Run KaptsjaHTMLpages.py." % (gen_failure_page, html_dir, gen_failure_page_file)
    logging.error(text)
    sys.exit(syserrmsg)
if os.path.exists(gen_success_page_file) != True:
    text = "File %s does not exist in directory %s.\nLooking for %s.\nCheck html_dir setting and spelling of file name (upper/lower case sensitive) or create an own success page with the correct name. Run KaptsjaHTMLpages.py." % (gen_success_page, html_dir, gen_success_page_file)
    logging.error(text)
    sys.exit(syserrmsg)
if not 'copyright_text_picture' in globals(): 
    copyright_text_picture = " "
if  type(copyright_text_picture) != str:
    text = "copyright_text_picture %s is not a string or no text has been defined. Nothing will be displayed." % copyright_text_picture
    logging.info(text)   
if not 'copyright_text_position' in globals(): 
    copyright_text_position = "R"
if copyright_text_position not in  ("R", "L", "r", "l"):
    copyright_text_position = "R"
if not 'textzone_center_text' in globals(): 
    textzone_center_text = True
if not type(textzone_center_text) == bool: 
    textzone_center_text = True
if not type(rotate_char_circle) == bool: 
    rotate_char_circle = True
if not type(rotate_char_circle) == int:
    rotate_angle_char_circle = 30
if -1 <= int(rotate_angle_char_circle) <= 361 : # when true we skip else
    pass
else:
    rotate_angle_char_circle = 30
try:
    import lxml
    beautifulsoup_parser = "lxml"
    text = "BeautifulSoup uses lxml parser"
    logging.info(text)
except:
    beautifulsoup_parser = "html.parser"
    text = "BeautifulSoup uses html.parser"
    logging.info(text)
    #
######################################
### Do not change any text literals here, but only in KaptsjaConfiguration.py!
######################################    
# Defines text literals in English when not set in Configuration or deleted by accident
# When text literal is set to None, no text will be shown
# When text literal is set to "", the default English text will be shown
# Else the "translated text" set in Configuration will be shown
text = "Please check box to continue3"
if not "check_the_box_text" in globals():
    check_the_box_text = text
if check_the_box_text == "":
    check_the_box_text = text
if check_the_box_text is None:
    check_the_box_text = ""

text = "Number of clicks:"
if not "number_of_clicks" in globals():
    number_of_clicks = text
if number_of_clicks == "":
    number_of_clicks = text
if number_of_clicks is None:
    number_of_clicks = ""

text = "Too many clicks. Retry."
if not "too_many_clicks_text" in globals():
    too_many_clicks_text = text
if too_many_clicks_text == "":
    too_many_clicks_text = text
if too_many_clicks_text is None:
    too_many_clicks_text = ""

text = "Clicking the exact number of circles enables Submit button"
if not "exact_nr_circles_submit" in globals():
    exact_nr_circles_submit = text
if exact_nr_circles_submit == "":
    exact_nr_circles_submit = text
if exact_nr_circles_submit is None:
    exact_nr_circles_submit = ""
    
text = "Submitted. Click button New Picture."
if not "submitted_text" in globals():
    submitted_text = text
if submitted_text == "":
    submitted_text = text
if submitted_text is None:
    submitted_text = ""

text = "Too many retries. Click button New Picture."
if not "too_many_retries_text" in globals():
    too_many_retries_text = text
if too_many_retries_text == "":
    too_many_retries_text = text
if too_many_retries_text is None:
    too_many_retries_text = ""

text = "New Picture"
if not "new_pic_button_text" in globals():
    new_pic_button_text = text
if new_pic_button_text == "":
    new_pic_button_text = text
if new_pic_button_text is None:
    new_pic_button_text = ""

text = "Retry"
if not "retry_button_text" in globals():
    retry_button_text = text
if retry_button_text == "":
    retry_button_text = text
if retry_button_text is None:
    retry_button_text = ""

text = "Submit Result"
if not "submit_button_text" in globals():
    submit_button_text = text
if submit_button_text == "":
    submit_button_text = text
if submit_button_text is None:
    submit_button_text = ""

text = "CaptCha for Human recognition"
if not "intro_captcha_text" in globals():
    intro_captcha_text = text
if intro_captcha_text == "":
    intro_captcha_text = text
if intro_captcha_text is None:
    intro_captcha_text = ""

text = "Human Verification"
if not "captcha_hdr_text" in globals():
    captcha_hdr_text = text
if captcha_hdr_text == "":
    captcha_hdr_text = text
if captcha_hdr_text is None:
    captcha_hdr_text = ""
    
# Close modal window character X or use another HTML entity defined as &xxx; 
# Symbol &times; equals to &#215;  text may be used style is H2
text = "&otimes;" 
if not "captcha_hdr_close" in globals():
    captcha_hdr_close = text
if captcha_hdr_close == "":
    captcha_hdr_close = text
if captcha_hdr_close is None:
    captcha_hdr_close = ""

text = "I am a not a robot&nbsp;&nbsp;"
if not "not_a_robot_text" in globals():
    not_a_robot_text = text
if not_a_robot_text == "":
    not_a_robot_text = text
if not_a_robot_text is None:
    not_a_robot_text = ""

text = "#a3ff9e"
if not "submit_button_bgcolor_enabled" in globals():
    submit_button_bgcolor_enabled = text
if submit_button_bgcolor_enabled == "":
    submit_button_bgcolor_enabled = text
if submit_button_bgcolor_enabled is None:
    submit_button_bgcolor_enabled = ""
    
text = "#ff9e9e"
if not "submit_button_bgcolor_disabled" in globals():
    submit_button_bgcolor_disabled = text
if submit_button_bgcolor_disabled == "":
    submit_button_bgcolor_disabled = text
if submit_button_bgcolor_disabled is None:
    submit_button_bgcolor_disabled = ""

text = "#a3ff9e"
if not "retry_button_bgcolor_enabled" in globals():
    retry_button_bgcolor_enabled = text
if retry_button_bgcolor_enabled == "":
    retry_button_bgcolor_enabled = text
if retry_button_bgcolor_enabled is None:
    retry_button_bgcolor_enabled = ""

text = "#ff9e9e"
if not "retry_button_bgcolor_disabled" in globals():
    retry_button_bgcolor_disabled = text
if retry_button_bgcolor_disabled == "":
    retry_button_bgcolor_disabled = text
if retry_button_bgcolor_disabled is None:
    retry_button_bgcolor_disabled = ""

text = "#a3ff9e"
if not "new_pic_button_bgcolor_enabled" in globals():
    new_pic_button_bgcolor_enabled = text
if new_pic_button_bgcolor_enabled == "":
    new_pic_button_bgcolor_enabled = text
if new_pic_button_bgcolor_enabled is None:
    new_pic_button_bgcolor_enabled = ""

text = "#ff9e9e"
if not "new_pic_button_bgcolor_disabled" in globals():
    new_pic_button_bgcolor_disabled = text
if new_pic_button_bgcolor_disabled == "":
    new_pic_button_bgcolor_disabled = text
if new_pic_button_bgcolor_disabled is None:
    new_pic_button_bgcolor_disabled = ""

text = "Are you human? Click in each circle. First Digits then Letters.\n In ASCENDING order (▲).\n Not readable? Click on New Picture button."
if not "text_Ascending" in globals():
    text_Ascending = text
if text_Ascending == "":
    text_Ascending = text
if text_Ascending is None:
    text_Ascending = ""
    
text = "Are you human? Click in each circle. First Letters then Digits.\n In DESCENDING order (▼).\n Not readable? Click on New Picture button."
if not "text_Descending" in globals():
    text_Descending = text
if text_Descending == "":
    text_Descending = text
if text_Descending is None:
    text_Descending = ""

if not "ico_file" in globals():
   ico_file = media_dir + r"Kaptsja.ico" 
if os.path.exists(ico_file) == False:
    text = "File %s did not exist in directory %s.\nIt will be created." % (ico_file, media_dir)
    logging.error(text)
    ico_encoded = b'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'
    try:
        ico_bin = b64decode(ico_encoded)
        f = open(ico_file, "wb")
        f.write(ico_bin)
        f.close()
    except:
        text = "File %s in directory %s.\nFailure to create. Is import base64 missing? Check ico_encoded in KaptsjaGenerator.py. Maybe corrupt?" % (ico_file, media_dir)
        logging.error(text)

##########################################
# END of validation of the configuration
##########################################

# create the Encryption Cipher object using the secret_key_file
# when a secret_key_file does not exists it will be created automatically in key_dir
cripto = Cripto(secret_key_file)


class Kaptsja(object):
    def __init__(self, plainstring="", mode=captcha_page_url):
        self.plainstring = plainstring
        self.charlist = []
        self.reverse = "0"
        self.randomlist = randomlist
        ##print("self.randomlist=", self.randomlist)
        self.picture_list = []
        self.mode = mode

    def rcolor(self):
        return '#{:06x}'.format(random.randint(0x011110, 0xFFFFFF))
        #return '#{:06x}'.format(random.randint(0, 256**3))  # alternative 1
        # return f"#{random.randrange(0x010000):06x}"         # alternative 2

    def rgb2hex (self, r,g,b):
        return tuple(lambda r,g,b: f"#{r:02x}{g:02x}{b:02x}")  # create tuple with 3 RGB digits add leading zeros when needed 
        
    def hex2rgb (self, hx):
        # hexcolor = hx.lstrip("#")
        return tuple(int(hx.lstrip("#")[i:i+2], 16) for i in (0, 2, 4)) # create tuple with hexadecimal color code with leading #
        
    def hex2rgb_transparent (self, hx, color_transparency):  # create tuple with 4 RGB digits add leading zeros when needed, last digit is transparency digit
        # hexcolor = hx.lstrip("#")
        T1 = tuple(int(hx.lstrip("#")[i:i+2], 16) for i in (0, 2, 4)) 
        # tuples are immutable, so we make a list and add extra transparancy value (128 means half transparent)
        L1 = list(T1)
        L1.append(int((100-color_transparency)*255/100))
        T1=tuple(L1)
        color = T1
        return color 

    def inverse_rgbcolor(self, rgbcolor):
        return tuple(map(lambda x: 255 - x, rgbcolor))
        
    # Calculate with Pythagoras formula the distance between 2 circle centers x1, y1 and x2, y2 
    # square root of (square X multiplied with Square Y) or in formula:√(x²+ y²) = √r² returns d (istance)
    def distance(self, x1, y1, x2, y2):
        dsq = (x1 - x2) ** 2 + (y1 - y2) ** 2
        d = dsq**(1/2)  # square root
        return d

    def text_wrap(self, text, font, max_width):
            # Wrap text into the text zone width to be display
            lines = []
            
            # When text width is smaller than the max_width (of text zone) no split is needed
            # Text is just added to the list with line list
            if font.getsize(text)[0]  <= max_width:
                lines.append(text)
            else:
                # "\n" in text forces a new line break so split text in multiple lines
                newtextlines = text.split('\n')
                for newtextline in newtextlines:
                    newtextline = newtextline.lstrip(" ") # strip leading spaces at start of a new line
                    # split the line by spaces into list of words
                    words = newtextline.split(' ')
                    i = 0  # index to count the added words in "display" line
                    # append each word to a "display" line as long as it fits (width is shorter than the textzone width)
                    while i < len(words):
                        line = '' # start with empty "display" line
                        while i < len(words) and font.getsize(line + words[i])[0] <= max_width: 
                            # check if there are still words to be added and if the width of the line + word to be added still fit in the text zone using the given font size width.
                            line = line + words[i]+ " " # add word to line plus a space
                            i += 1  # update counter
                        if not line: # in case there
                            line = words[i]
                            i += 1
                        lines.append(line)
            return lines

    def create_unique_file(self, unique_time, infile):
        infile_name, infile_ext = os.path.splitext(infile)
        generated_unique_infile = infile_name + "_" + self.unique_time + infile_ext
        generated_unique_outfile     = work_dir + generated_unique_infile
        generated_unique_outfile_url = work_url + generated_unique_infile
        return generated_unique_outfile, generated_unique_outfile_url

    def create_all_filenames_and_fileurls(self):
        self.unique_time = str(time.time())
        # The file names and urls are generated here
        self.gen_captcha_page_file, self.gen_captcha_page_file_url = self.create_unique_file(self.unique_time, gen_captcha_page) # first is the file, second is url
        self.generated_unique_picture_file, self.generated_unique_picture_file_url = self.create_unique_file(self.unique_time, gen_picture) # first is the file, second is url
        self.gen_div_page_file, self.gen_div_page_file_url = self.create_unique_file(self.unique_time, gen_div_page) # first is the file, second is url
        self.gen_modal_page_file, self.gen_modal_page_file_url = self.create_unique_file(self.unique_time, gen_modal_page) # first is the file, second is url
        self.gen_css_file, self.gen_css_file_url = self.create_unique_file(self.unique_time, gen_css) # first is the file, second is url
        self.gen_javascript_file, self.gen_javascript_file_url = self.create_unique_file(self.unique_time, gen_javascript) # first is the file, second is url

    def set_charlist(self):
        if self.plainstring == None or self.plainstring == " ":
            self.charlist = random.sample(used_letters, number_of_letters)+ random.sample(used_digits, number_of_digits)
            logging.info(r"Create Kaptsja: charlist generated: %s" % self.charlist)
        else:
            self.charlist = list(self.plainstring)   # use the provided characters given to this function
            logging.info(r"Create Kaptsja: charlist provided: %s" % self.charlist)
        random.shuffle(self.charlist)
        self.reverse = random.choice(["0", "1"])
        # Return two values: A checkvalue (charlist) which contains the shuffled and previously entered 6 digits or characters
        # An encrypted controlvalue created from the checkvalue.
        # The controlvalue should be send hidden along with the Kaptsja map to the user.
        # The user clicks on the Kaptsja map in the correct way and returns the clicked result including the controlvalue.
        # The Encryption function expects a single string input. Therefore the numeric digits in charlist 
        # are first converted into string digits and then joined into a single string 
        self.plainstring = ''.join(map(str, self.charlist))
        self.encodedstring = cripto.encryption(self.plainstring)
        self.plainstring
        return 0
        
    def get_filelist(self, dir, filter="*"):
        filelist =[]
        for filename in glob.glob(randomlist_dir +'*'):
            infile_name, infile_ext = os.path.splitext(filename)
            if infile_ext.lower() in allowed_file_extensions:
                filelist.append(filename)
            ##print(dir, filter, filename)
        return filelist
        
    def create_picturelist(self):
        if self.randomlist:
            self.picture_list = self.get_filelist(randomlist_dir, filter="*")
            ##print("self.picture_list =", self.picture_list )
        else:
            self.picture_list.append(os.path.normpath(input_picture_file))

    def set_random_picture(self):
        self.input_picture = random.choice(self.picture_list) 
        ##print("self.input_picture=",self.input_picture)

    def get_font(self, font_type, font_points, text):
        fnt = ImageFont.truetype(font_type, font_points)
        fw = 0
        fh = 0    
        fw_max = 0 
        fh_max = 0 
        for char in text:
            fw, fh = fnt.getsize(text = char)
            if fw > fw_max: fw_max = fw
            if fh > fh_max: fh_max = fh
        return fnt, fw_max, fh_max
        
    def scale_input_picture(self):
        self.set_random_picture()
        picture = Image.open(self.input_picture)
        # When closing the Bottle server with CTRL-C this error might appear: 
        # sys:1: ResourceWarning: unclosed file <_io.BufferedReader name='./<dir>/<name>.jpg'>
        # The reason is that Image.open function keeps the file open till picture is closed or saved.
        # This picture must be opened here but it is not (yet) closed when CTRL-C  is used!   
        # Get Width and height of the Input image 
        self.wpic, self.hpic = picture.size
        logging.info(r"Create Kaptsja: picture %s provided with size (w, h): %s" %(picture.filename, picture.size) )
        rescale = False  # default no rescale, unless...
        #  all input pictures will be automatically rescaled once and original will be overwritten!
        if input_picture_rescale_width > 0:       
            # input picture will be rescaled when value is between min and max values (see validation)
            rescale_wpic = input_picture_rescale_width
            rescale      = True
        if self.wpic > 598: 
            rescale_wpic = 598  # This is the maximum allowed width (forced as default maximum of bootstrap modal is 600px)
            rescale      = True
        if rescale == True:
            factor = (rescale_wpic / self.wpic)
            rescale_hpic = int((self.hpic * factor)) # same percentage of resizing as width must be integer not float
            text = "Provided image width %s exceeds the max width of the modal window size which is 600px.\nRezing and saving image %s to %s x %s pixels" % (picture, self.wpic, rescale_wpic, rescale_hpic)
            logging.info(text)     
            picture = picture.resize((rescale_wpic,rescale_hpic), resample=0)
            self.wpic, self.hpic = picture.size
            self.picture = picture
            picture.save(self.input_picture)
        # make a blank image from input picture (might be resized!) to allow opacity with transparency drawings
        self.picRGBA = picture.convert('RGBA')
        self.img = Image.new('RGBA', self.picRGBA.size, (255,255,255,0))
        # Create a Draw object to put text, circles etc on
        self.draw = ImageDraw.Draw(self.img)  
        return 0

    def create_textzone(self):
        # Get Font for text zone 
        fnt_for_textzone, fwt_max, fht_max = self.get_font(font_textzone, font_pnts_textzone, list(text_Ascending+text_Descending))
        # set text for Ascending / Descending
        if self.reverse == "0":  # "0" = Ascending
            text=text_Ascending
        else:                    # "1" = Descending
            text=text_Descending
        # calculate the text zone positioning, based on the number of lines and wrap the text properly so it fits within the given boundaries.    
        text_x_min = border_line    # left start x-position for a text line (in the textzone)
        text_x_max = self.wpic - border_line  # right end x-position for a text line (in the textzone)
        text_max_width = text_x_max - text_x_min
        lines = self.text_wrap(text, fnt_for_textzone, text_max_width)
        line_height = fht_max # a line height is determined by the maximum height of the font
        text_y_min  = self.hpic - (fht_max * len(lines)) # top start y-position for a first text line
        text_y_max  = self.hpic - border_line  # bottom end y-position for the last text line (in the textzone)
        text_y_max  = text_y_max - (len(lines)*line_height) # adjusted y-position. 
        x = text_x_min
        y = text_y_max
        # draw rectangle at bottom, used transparency as defined in configuration setting
        textzone_fill_color_transparent = self.hex2rgb_transparent(textzone_bg_color, textzone_color_trans)
        self.draw.rectangle((0,text_y_min - border_line, self.wpic, self.hpic - border_line), fill=textzone_fill_color_transparent) #, outline = line_color, width=0)
        # position aligned text line by line 
        for line in lines:
            line_fnt_width = fnt_for_textzone.getsize(line)[0]
            if textzone_center_text == True:
                self.draw.text(( text_max_width//2 - line_fnt_width//2   ,y), line, fill=textzone_text_color, font=fnt_for_textzone)
            else:
                self.draw.text((x,y), line, fill=textzone_text_color, font=fnt_for_textzone)
            y = y + line_height    # update y-axis for new line
         # draw top en bottom border_circle of the rectangle, no left, right border_circle    
        self.draw.line((0, text_y_min - border_line, self.wpic, text_y_min  - border_line), fill=line_color, width=border_line)
        self.draw.line((0, self.hpic - border_line//2, self.wpic, self.hpic - border_line//2), fill=line_color, width=border_line)
        
        # draw lines over text to morph the text a little
        for line_h in range (text_y_min  - border_line, self.hpic, 10):
            self.draw.line((0,line_h, self.wpic, line_h), fill=line_color, width=1)
        self.textzone_y_top = text_y_min - border_line//2 # Circles may not be drawn in the textzone and its border_circle  
        return 0
        
    def create_copyright_text(self, text):
        # Position the Copyright Text        
        fnt_for_copyright, copyright_text_width, copyright_text_height = self.get_font(font_textzone, font_pnts_copyright_text, text)
        copyright_text_width = fnt_for_copyright.getsize(copyright_text_picture)[0]
        if copyright_text_position.lower() == "r":
            self.draw.text((self.wpic - copyright_text_width - 5, 5), copyright_text_picture, fill=copyright_text_color, font=fnt_for_copyright)
        else: # left upper corner
            self.draw.text((5, 5), copyright_text_picture, fill=copyright_text_color, font=fnt_for_copyright)
        return 0
        
    def create_circles(self):
        # Get font for circles
        fnt_for_circle, fwc_max, fhc_max = self.get_font(font_circle, font_pnts_circle, self.charlist)
        # circle dimensions depends on max size of fw_max anf hh_max
        if fwc_max > fhc_max:
            cdiam = fwc_max * 2 # cdiam circle diameter excluding a border_circle size, 2 times the max width of selected font
        else:
            cdiam = fhc_max * 2 # cdiam circle diameter excluding a border_circle size, 2 times the max height of selected font
        rad_minimum = cdiam // 3 # the minimum radius of a circle
        # Create a list of mapareas: used in HTML file, maparea's must be "drawn by browser exactly on top of circles in picture")
        self.map_area_list = []
        # Create a list of circles. Each circle is a tuple with 3 components, the x and y coordinate and the radius. Append new circles as long the length of the list is less than n:
        circle_list = []
        # i is index for identify the ID of the HTML area's and when clicked ID is used to create a checkvalue in the HTML form
        i = 0  
        # Counter for the number of evaluations a
        # The allowed_evals_per_char forces a break to prevent an endless loop 
        # in case not all circles will fit on the given space/pictures. 
        evaluations = 0
        max_evaluations = len(self.charlist) * allowed_evals_per_char
        while len(circle_list) < len(self.charlist):
            evaluations = evaluations + 1
            if evaluations > max_evaluations: # evaluation stopped as there is most likely no place on picture anymore
                logging.error(r"Create Kaptsja: Too high number of Circle collisions therefore a break is forced.")
                logging.error(r"Create Kaptsja: Check picture size, font size, given number of characters and adapt their values.")
                logging.error(r"Create Kaptsja: (Width, Height) of picture:%s and font %s, number of characters: %s" %((self.wpic, self.hpic), (fwc_max, fhc_max), len(self.charlist)))
                logging.error(r"Create Kaptsja:   1. Start with reducing the number of characters on the Kaptsja picture.")
                logging.error(r"Create Kaptsja:   2. Reduce font size.")
                pic_minheight = self.hpic - self.textzone_y_top + (cdiam + border_circle+border_line)*3  # recommended height for a picture to avoid too much collisions
                logging.error(r"Create Kaptsja:   3. Increase picture size (Recommended minimum height is %s)." % pic_minheight)
                return 1
     
            try:
                #Create a random x, y position and radius for circles:
                r = rad_minimum # sets the minimum radius
                r = random.randint(r, r + 15)
                x = random.randint(0, self.wpic - (r + border_circle) * 2 )
                y = random.randint(0, self.textzone_y_top - (r + border_circle) * 2)
                d = (r+border_circle) * 2
                char= self.charlist[i]
            except ValueError as e:
                logging.error(r"Create Kaptsja: Text zone doesn't fit on this picture. Calculated value is %s." % self.textzone_y_top)
                logging.error(r"Create Kaptsja: Check picture size, font size, given number of characters and adapt their values.")
                logging.error(r"Create Kaptsja: (Width, Height) of picture:%s and font %s, number of characters: %s" %((self.wpic, self.hpic), (fwc_max, fhc_max), len(self.charlist)))
                logging.error(r"Create Kaptsja:   1. Start with reducing the number of characters on the Kaptsja picture.")
                logging.error(r"Create Kaptsja:   2. Reduce font size.")
                pic_minheight =  self.hpic - self.textzone_y_top + (cdiam + border_circle)*3   # recommended height for a picture to avoid too much collisions
                logging.error(r"Create Kaptsja:   3. Increase picture size (Recommended minimum height is %s)." % pic_minheight)
                return 1
            # Evaluate if the circle intersects with an other circle which is in the list
            collide = False
            # # use Pythagoras formula to calculate distances between new circle and the ones in the circle_list.
            for x2, y2, r2 in circle_list:
                # add each evaluations to counter
                evaluations = evaluations + 1
                d = self.distance(x, y, x2, y2)
                if d < r*1.9//1 + r2 :
                    collide = True
                    break
            # Append the circle to list if it does not collide:
            if not collide:
                circle_list.append((x, y, r))
                # Draw the circle on the image and place character in circle
                circle_fill_color = self.rcolor()
                circle_fill_color_transparent = self.hex2rgb_transparent(circle_fill_color, circle_color_trans)
                fw, fh = self.draw.textsize(text="{0}".format(char),font=fnt_for_circle) 
                # text_color is the inversed value color of circle_fill_color which is a hexcode color
                # Below the hexcode color is converted to rgb code and then change to it's inversed value
                text_color = self.inverse_rgbcolor(self.hex2rgb_transparent( circle_fill_color, char_circle_color_trans))

                # Use ONLY anchor in draw.text (....., anchor="mm") when PIL version is 8.0.0 or higher!
                # The anchor parameter was present in earlier versions of Pillow, but implemented only in version 8.0.0. The anchor parameter is ignored for non-TrueType fonts.
                if rotate_char_circle == True: 
                    # Create a transparent image mask with sizes of a (circle radius + border_circle) * 2 , that's a square 
                    img_mask = Image.new('RGBA', (r * 2+border_circle, r * 2+border_circle), color=(0,0,0,0) )
                    draw_img_mask = ImageDraw.Draw(img_mask)
                    # draw the circle on this square mask and fill circle
                    draw_img_mask.ellipse(((0,0), r * 2, r * 2), fill=circle_fill_color_transparent, outline=self.rcolor(), width=border_circle)
                    # Create a second image with same size as the mask
                    img_char = Image.new('RGBA', (r * 2+border_circle, r * 2+border_circle), color = circle_fill_color_transparent)
                    draw_img_char = ImageDraw.Draw(img_char)
                    # pick a random angle to rotate this second image once the character has been drawn
                    angle = random.randint(int(rotate_angle_char_circle) * -1, int(rotate_angle_char_circle))
                    # calculate the xy position of the character in the circle
                    xy=(r - (fw / 2) , r - (fh / 2))
                    # draw the character on second image
                    draw_img_char.text(xy=(r - (fw / 2) , r - (fh / 2)), text="{0}".format(char), fill=text_color,font=fnt_for_circle)
                        
                    # self.draw_rotated_text(img_char, angle, xy, text="{0}".format(char), fill=text_color,font=fnt_for_circle)
                    # morphing the character with pie slices positioned from the circle center. 
                    # the size of the pie slice is randomly. 
                    num_of_pieslice = random.randint(3, max_morphing_pie_slices)  
                    pieslice_start_angles = random.sample(range(0,360), num_of_pieslice)
                    for pieslice_start_angle in pieslice_start_angles:
                        pieslice_end__angle = pieslice_start_angle+(random.randint(3,15) )
                        draw_img_char.pieslice((border_circle * 2, border_circle * 2, (r - border_circle) * 2 , (r - border_circle) * 2), pieslice_start_angle, pieslice_end__angle, fill=circle_fill_color_transparent)
                    # put a line under numbers 6 and 9 to avoid recognition mistakes when characters are rotated "too" much
                    if char in ("6", "9"):
                        draw_img_char.line((r - (fw / 2), r - (fh / 2) + fh, r - (fw / 2) + fw, r - (fh / 2) + fh), fill=text_color, width=1)
                    img_char = img_char.rotate(angle)

                    self.img.paste(img_char, (x,y), img_mask)
                    # angle = random.randint(int(rotate_angle_char_circle) * -1, int(rotate_angle_char_circle))
                    # xy=(x + r - (fw // 2) , y + r - (fh // 2))
                    # self.draw_rotated_text(self.img, angle, xy, text="{0}".format(char), fill=text_color,font=fnt_for_circle)
                else:
                    self.draw.ellipse(((x,y), x + r * 2, y + r * 2), fill=circle_fill_color_transparent, outline=self.rcolor(), width=border_circle)
                    self.draw.text(xy=(x + r - (fw // 2) , y + r - (fh // 2)), text="{0}".format(char), fill=text_color,font=fnt_for_circle) 
                    # morphing the character with pie slices positioned from the circel center. 
                    # the size of the pie slice is randomly. 
                    num_of_pieslice = random.randint(3, max_morphing_pie_slices)  
                    pieslice_start_angles = random.sample(range(0,360), num_of_pieslice)
                    for pieslice_start_angle in pieslice_start_angles:
                        pieslice_end__angle = pieslice_start_angle+(random.randint(3,15) )
                        self.draw.pieslice((x + border_circle, y + border_circle, x + r * 2 - border_circle, y + r * 2 - border_circle), pieslice_start_angle, pieslice_end__angle, fill=circle_fill_color_transparent)
                # Append the area to maparea list:
                self.map_area_list.append('<area id="A%s" shape="circle" coords="%s, %s, %s" alt="Click here" title="Click area" onclick="fclick(%s);"></area>' % (random.randint(100,200), x + r, y + r, r + border_circle // 2, i) )
                i=i+1  # count this circle
        self.create_copyright_text(copyright_text_picture)
        out = Image.alpha_composite(self.picRGBA, self.img)
        out.save(self.generated_unique_picture_file)
        logging.info(r"Create Kaptsja: generated Kaptsja picture PNG file: %s" % self.generated_unique_picture_file)
        return 0

    # keep this method, however it is not in use 
    def draw_rotated_text(self, image, angle, xy, text, fill, *args, **kwargs):
        """ Draw text at an angle into an image, takes the same arguments
            as Image.text() except for:

        :param image: Image to write text into
        :param angle: Angle to write text at
        """
        # get the size of our image
        width, height = image.size
        max_dim = max(width, height)

        # build a transparency img_mask large enough to hold the text
        img_mask_size = (max_dim * 2, max_dim * 2)
        img_mask = Image.new('L', img_mask_size, 0)

        # add text to img_mask
        draw = ImageDraw.Draw(img_mask)
        draw.text((max_dim, max_dim), text, 255, *args, **kwargs)

        if angle % 90 == 0:
            # rotate by multiple of 90 deg is easier
            rotated_img_mask = img_mask.rotate(angle)
        else:
            # rotate an an enlarged img_mask to minimize jaggies
            bigger_img_mask = img_mask.resize((max_dim*8, max_dim*8),
                                      resample=Image.BICUBIC)
            rotated_img_mask = bigger_img_mask.rotate(angle).resize(
                img_mask_size, resample=Image.LANCZOS)

        # crop the img_mask to match image
        img_mask_xy = (max_dim - xy[0], max_dim - xy[1])
        b_box = img_mask_xy + (img_mask_xy[0] + width, img_mask_xy[1] + height)
        img_mask = rotated_img_mask.crop(b_box)

        # paste the appropriate color, with the text transparency img_mask
        color_image = Image.new('RGBA', image.size, fill)
        image.paste(color_image, img_mask)

                
# All braces for css and javascript itself MUST be double notation like {{ // some css or javascript code }} to escape the first. 
# Only one brace will be generated in the output. All formatting keys like .format(key= ) MUST have a single brace {key} notation.

    def html_header_css(self):
        return '''        
        //  this file/part is generated by KaptsjaGenerator.py       
        area {{
        cursor:pointer;
        }}
        table tr td {{
            border: 5px; 
            text-align: center; 
            vertical-align: middle;
        }}
        #count_clicks_per_try {{
            border:1px;
        }}
        #count_max_retries {{
            border:1px; 
        }}
        #message {{
            border:1px; 
            color: red; 
        }}
        #submitbutton[disabled]
            {{
             background: {retry_button_bgcolor_disabled};  # redish
            }}
        #submitbutton
            {{
             background: {retry_button_bgcolor_enabled}; // greenish
            }}
        #retrybutton[disabled]
            {{
             background: {retry_button_bgcolor_disabled};  # redish
            }}
        #retrybutton
            {{
             background: {retry_button_bgcolor_enabled}; // greenish
            }}
        #reloadbutton[disabled]
            {{
             background: {new_pic_button_bgcolor_disabled};  # redish
            }}
        #reloadbutton
            {{
             background: {new_pic_button_bgcolor_enabled}; // greenish
            }}
        '''.format(submit_button_bgcolor_enabled=submit_button_bgcolor_enabled, \
            submit_button_bgcolor_disabled=submit_button_bgcolor_disabled, \
            retry_button_bgcolor_enabled=retry_button_bgcolor_enabled, \
            retry_button_bgcolor_disabled=retry_button_bgcolor_disabled, \
            new_pic_button_bgcolor_enabled=new_pic_button_bgcolor_enabled, \
            new_pic_button_bgcolor_disabled=new_pic_button_bgcolor_disabled, \
            )
# modal
    def html_header_css_modal(self):
        return ''' 
        {html_header_css}
        .close:focus, button.close:hover {{
            color: #FF0000;
        }}
        .modal-header {{
            padding-left: 15px;
            padding-right: 15px;
            padding-top: 2px;
            padding-bottom: 2px;
        }}
        .modal-body {{
            padding-left: 0px;
            padding-right: 0px;
            padding-top: 2px;
            padding-bottom: 2px;
        }}
        .modal-footer {{
            padding-left: 15px;
            padding-right: 15px;
            padding-top: 2px;
            padding-bottom: 2px;
        }}
        .modal-content {{
        width:{width}px;
        position: relative;
        background-color: #fff;
        -webkit-background-clip: padding-box;
        background-clip: padding-box;
        border: 1px solid #999;
        border: 1px solid rgba(0,0,0,.2);
        border-radius: 6px;
        outline: 0;
        -webkit-box-shadow: 0 3px 9px rgba(0,0,0,.5);
        box-shadow: 0 3px 9px rgba(0,0,0,.5);
        }}
    `   '''.format(html_header_css=self.html_header_css(), \
            width=self.wpic+2, \
            ) 
        # 2 pixels extra are needed to get good right alignment (generated max picture width is 2 pixels than max widt of modal-content
    def javascript_window_onload_modal(self):
        return ''' '''
    def html_header_modal(self):
        return '''<div class="modal-header">
                <button type="button" class="close" data-dismiss="modal" aria-hidden="true"><h2>{captcha_hdr_close}</h2></button>
                <h3>{captcha_hdr_text}</h3>
                </div>
        '''.format(captcha_hdr_close=captcha_hdr_close, \
            captcha_hdr_text=captcha_hdr_text)
    def html_body_modal(self):
        return '''<div class="modal-body text-center">
                </div>
        '''
    def html_footer_modal(self):
        return '''<!--  OK button not used; uncomment if you want this button to appear  -->
                <div class="modal-footer">
               <!-- <button type="button" class="btn btn-primary" data-dismiss="modal">OK</button>  -->
                </div> 
                <!-- Include first the jQuery and second Bootstrap javascript libraries. Order is important!  
                <script src="{jquery_js_file}"></script>
                <script src="{bootstrap_js_file}"></script> -->
        '''.format (jquery_js_file=jquery_js_file, bootstrap_js_file=bootstrap_js_file)
    def javascript_freload_modal(self):
        return '''
        function freload() {{
            setTimeout(function (){{
            // Delay time of 1.5 seconds between clicks on New Picture button will help to delete meanwhile chached Kaptsja's
            }}, 1500); // How long do you want the delay to be (in milliseconds)? 
        
            // following is required to force browser to reload always the page!
            var tnow = new Date(Date.now());
            var formatted_stamp = tnow.getHours() + "_" + tnow.getMinutes() + "_" + tnow.getSeconds() + "_" +  Math.random() * 1000;
            // Following line shows modal window direct, next block used fancy fade in and out, comment this line and uncomment the block is you want to use it
            $("#myModal").find(".modal-content").load("{modal_captcha_page_url}?random=" + formatted_stamp ).show();
            //  Show modal with fade in and fade out effect This is an alternative for the line above
            //  $('#myModal').fadeOut("slow",function(){{
            //        // specify here actions you need
            //        }}).fadeIn("slow",function(){{
            //            $("#myModal").find(".modal-content").load("{modal_captcha_page_url}?random=" + formatted_stamp ).show();
            //        }});                                               
        }}
        '''.format(modal_captcha_page_url=modal_captcha_page_url )
    def html_reload_button_modal(self):nu
        return '''\n                        <input type="button" id="reloadbutton" name="reloadbutton" value="{new_pic_button_text}" onclick = "freload();"></input> 
        '''.format(new_pic_button_text=new_pic_button_text)
# div
    def javascript_window_onload_div(self):
        return ''' '''
    def javascript_freload_div(self):
        return '''
        function freload (e) {{
            setTimeout(function (){{ 
            // Delay time of 1.5 seconds between clicks on New Picture button will help to delete meanwhile chached Kaptsja's
            }}, 1500); // How long do you want the delay to be (in milliseconds)? 
        
        // following is required to force browser to reload always the page!
            var tnow = new Date(Date.now());
			var formatted_stamp = tnow.getHours() + "_" + tnow.getMinutes() + "_" + tnow.getSeconds() + "_" +  Math.random() * 1000;

            (e || window.event).preventDefault();
            var con = document.getElementById('content')
            ,   xhr = new XMLHttpRequest();
            xhr.onreadystatechange = function (e) {{ 
                if (xhr.readyState == 4 && xhr.status == 200) {{
                 con.innerHTML = xhr.responseText;
                }}
        }}
        xhr.open("GET", "{div_captcha_page_url}?random=" +  formatted_stamp, true);
        xhr.setRequestHeader('Content-type', 'text/html');
        xhr.send();
        }}
        '''.format(div_captcha_page_url=div_captcha_page_url )
    def html_reload_button_div(self):
        return '''\n                        <input type="button" id="reloadbutton" name="reloadbutton" value="{new_pic_button_text}" onclick = "freload();"> </input> 
        '''.format(new_pic_button_text=new_pic_button_text)
# default
    def javascript_window_onload_default(self):
        return '''
        window.onload = function()  {
             document.getElementById('checkvalue').value = "";
             document.getElementById('count_clicks_per_try').value = "0";
             document.getElementById('submitbutton').disabled = true;
             document.getElementById("count_max_retries").value = "0";
              }
        '''
    def javascript_freload_default(self):
        return '''
        function freload() {{
            setTimeout(function (){{ 
            // Delay time of 1.5 seconds between clicks on New Picture button will help to delete meanwhile chached Kaptsja's
            }}, 1500); // How long do you want the delay to be (in milliseconds)? 
            location.reload(true);             
        }}
        '''   
    def html_reload_button_default(self):
        return '''\n                        <input type="button" id="reloadbutton" name="reloadbutton" value="{new_pic_button_text}" onclick = "freload();"> </input> 
        '''.format(new_pic_button_text=new_pic_button_text)
    def javascript_fretry_default(self):
        return '''
        function fretry()  {{
          if ( document.getElementById("count_clicks_per_try").value > "0") {{ 
            document.getElementById("count_max_retries").value++;
            }}
          document.querySelector("#message > span").innerHTML = "";
          document.getElementById('checkvalue').value = "";
          document.getElementById('count_clicks_per_try').value = "0";
          document.getElementById('submitbutton').disabled = true;
          if (document.getElementById("count_max_retries").value > {max_retries}) {{
                document.querySelector("#message > span").innerHTML = "{too_many_retries_text}";
                document.getElementById('submitbutton').disabled = true;
                document.getElementById('retrybutton').disabled = true;
                // document.getElementById("myForm").submit();        
                return false;        
            }}        
         }} 
        '''.format(max_retries=max_retries, \
            too_many_retries_text=too_many_retries_text)
    def javascript_fclick_default(self):
        return ''' 
        function fclick(value_in) {{
            if (document.querySelector("#message > span").innerHTML == "{submitted_text}")
            {{
                document.getElementById('submitbutton').disabled = true;
                document.getElementById('retrybutton').disabled = true;
                return false;
            }}
            if (document.querySelector("#message > span").innerHTML == "{too_many_retries_text}")
            {{
                document.getElementById('submitbutton').disabled = true;
                document.getElementById('retrybutton').disabled = true;
                return false;
            }}
            document.getElementById("count_clicks_per_try").value++;
            document.getElementById("checkvalue").value =                   document.getElementById("checkvalue").value.concat(value_in);
            if (document.getElementById("count_clicks_per_try").value > {max_clicks_per_try} - 1 )  {{
                document.getElementById('submitbutton').disabled = false;
                }}
            else 
                {{
                document.querySelector("#message > span").innerHTML = "";
                document.getElementById('submitbutton').disabled = true;
                }}
            
            if (document.getElementById("count_clicks_per_try").value > {max_clicks_per_try}) {{
                var retcode = fretry();
                if ( retcode != false ) 
                    {{
                    document.querySelector("#message > span").innerHTML = "{too_many_clicks_text}"; 
                    document.getElementById('submitbutton').disabled = true;
                    }}
                return false;  
            }}
        }}
        '''.format(max_clicks_per_try=number_of_digits+number_of_letters, \
            submitted_text=submitted_text, \
            too_many_clicks_text=too_many_clicks_text, \
            too_many_retries_text=too_many_retries_text ) 
    def javascript_fsubmit_default(self):
        return '''
        function fsubmit() {{
            document.querySelector("#message > span").innerHTML = "{submitted_text}"; 
            document.getElementById("myForm").submit();
            // by exceeding the maximum values a reload of new picture will be forced (to avoid back page in browser)
            document.getElementById("count_max_retries").value = {max_retries} + 1;
            document.getElementById("count_clicks_per_try").value = {max_clicks_per_try} + 1;             
        }}
        '''.format(max_retries=max_retries, max_clicks_per_try=number_of_digits+number_of_letters, \
        submitted_text=submitted_text) 
        
    def gen_javascript(self):
        if self.mode in (modal_page_url, modal_captcha_page_url): 
            javascript_window_onload = self.javascript_window_onload_modal()
            javascript_freload = self.javascript_freload_modal()
            javascript_fretry =self.javascript_fretry_default()
            javascript_fclick =self.javascript_fclick_default()
            javascript_fsubmit =self.javascript_fsubmit_default()
        if self.mode in (div_page_url, div_captcha_page_url): 
            javascript_window_onload = self.javascript_window_onload_div()
            javascript_freload = self.javascript_freload_div()
            javascript_fretry =self.javascript_fretry_default()
            javascript_fclick =self.javascript_fclick_default()
            javascript_fsubmit =self.javascript_fsubmit_default()
        if self.mode == captcha_page_url: 
            javascript_window_onload = self.javascript_window_onload_default()
            javascript_freload = self.javascript_freload_default()
            javascript_fretry =self.javascript_fretry_default()
            javascript_fclick =self.javascript_fclick_default()
            javascript_fsubmit =self.javascript_fsubmit_default()
        return '''//this file is generated by KaptsjaGenerator.py       
        {javascript_window_onload}
        {javascript_freload}
        {javascript_fretry}
        {javascript_fclick}
        {javascript_fsubmit}
        '''.format( javascript_window_onload=javascript_window_onload, \
                    javascript_freload=javascript_freload, \
                    javascript_fretry=javascript_fretry, \
                    javascript_fclick=javascript_fclick, \
                    javascript_fsubmit=javascript_fsubmit )  

    def modal_html(self):
        return '''<!doctype html>
    <html>
    <!-- this file is generated by KaptsjaGenerator.py -->
    <head>
    <meta charset="utf-8" />
    <title>Kaptsja</title>
    <!-- Include Bootstrap 3.3.7 CSS here in the head section -->
    <link rel="stylesheet" href="{bootstrap_css_file}">
    <style>
    .modal-content {{
        width:{modal_content_width}px;
        position: relative;
        background-color: #fff;
        -webkit-background-clip: padding-box;
        background-clip: padding-box;
        border: 1px solid #999;
        border: 1px solid rgba(0,0,0,.2);
        border-radius: 6px;
        outline: 0;
        -webkit-box-shadow: 0 3px 9px rgba(0,0,0,.5);
        box-shadow: 0 3px 9px rgba(0,0,0,.5);
    }}
    </style>
    <script>
    </script
    </head>
    <body>
    <!-- The data-toggle="modal" is necessary and should be copied verbatim to your code!
         The data-target="#modal" is also mandatory, but you should change #modal to #<ID of
         the div with class modal that contains your desired modal>.
         The value of the href attribute is the file Bootstrap will load HTML into the modal from. -->
    <div class="modal-header">
    <h3>{check_the_box_text}<h3>    
    <div> 
        <!-- When not using default settings you might need to adapt href in next line -->
        <a data-toggle="modal" data-target="#myModal" href="{modal_captcha_page_url}">
            <label>{not_a_robot_text}<input type="checkbox" name="checkbox" value="link"></label>
        </a>
    </div> 
    </div>     
    <!-- <a data-toggle="modal" data-target="#myModal" href="{modal_captcha_page_url}">{intro_captcha_text}</a> -->
    <!-- This entire nested div structure is necessary for modals -->
    <div class="modal fade text-center" id="myModal">
        <div class="modal-dialog">
            <div class="modal-content">
                <!-- 
                    The external HTML will be inserted here.
                     Note: you can just place the header-body-footer div structure in the
                     external file. For example: you could place the following in externalfile.html:
                <div class="modal-header">
                    <button type="button" class="close" data-dismiss="modal" aria-hidden="true">&times;</button>
                </div>
                <div class="modal-body">
                    <p>Hello Kaptsja!</p>
                </div>
                <div class="modal-footer">
                    <button type="button" class="btn btn-primary" data-dismiss="modal">OK</button>
                </div> 
                --> 
            </div>
        </div>
    </div>
    <!-- Include First the jQuery and Second Bootstrap javascript libraries. Order is important! -->
    <script src="{jquery_js_file}"></script>
    <script src="{bootstrap_js_file}"></script>
    </body>
    <html>
    '''.format(bootstrap_css_file=bootstrap_css_file, \
        bootstrap_js_file=bootstrap_js_file, \
        jquery_js_file=jquery_js_file, \
        modal_captcha_page_url=modal_captcha_page_url, \
        modal_content_width=input_picture_rescale_width + 2, \
        intro_captcha_text=intro_captcha_text, \
        not_a_robot_text=not_a_robot_text, \
        check_the_box_text=check_the_box_text )  
    # NOTE: 2 extra pixels are added to align picture nicely in the modal Window 
    # The max allowed width of picture is set to 598 as max default width for bootstrap modal is 600 pixels incl. 1px border for left and right )
    
    def div_html(self):
        return '''<!doctype html>
    <html>
    <!-- this file is generated by KaptsjaGenerator.py -->
    <head>
    <meta charset="utf-8" />
    <title>Kaptsja</title>
    <link rel="stylesheet" href="{gen_css_file_url}">
    <script src="{gen_javascript_file_url}">
    <script>
    </script>
    </head>
    <body>
    <div>
    <h1>{intro_captcha_text}<h1>
    </div>
    <table>
      <tr><td>
        <div id="content">
                <label>{not_a_robot_text}</label>
                <a href="{div_captcha_page_url}?random={force_reload}"><input type="checkbox" name="checkbox"  onclick="freload();">
            </a>
        </div>  
      </td></tr>
    </table>
    </body>
    </html>
        '''.format(gen_css_file_url=self.gen_css_file_url, \
        gen_javascript_file_url=self.gen_javascript_file_url, \
        captcha_page_url=captcha_page_url, \
        max_clicks_per_try=number_of_digits+number_of_letters, \
        max_retries=max_retries, \
        div_captcha_page_url=div_captcha_page_url, \
        force_reload=str(time.time()), \
        intro_captcha_text=intro_captcha_text, \
        not_a_robot_text=not_a_robot_text        )

    def captcha_html(self):
        if self.mode in (modal_page_url, modal_captcha_page_url): 
            html_header_css = self.html_header_css_modal()
            gen_javascript = self.gen_javascript()
            modal_header = self.html_header_modal()
            modal_body = self.html_body_modal()
            html_reload_button = self.html_reload_button_modal()
            modal_footer = self.html_footer_modal()
        if self.mode in(div_page_url, div_captcha_page_url): 
            html_header_css = self.html_header_css()
            gen_javascript = self.gen_javascript()
            modal_header = ""
            modal_body = ""
            html_reload_button = self.html_reload_button_div()
            modal_footer = ""
        if self.mode == captcha_page_url: 
            html_header_css = self.html_header_css()
            gen_javascript = self.gen_javascript()
            modal_header = ""
            modal_body = ""
            html_reload_button = self.html_reload_button_default()
            modal_footer = ""
        return '''<html>
            <!-- this file is generated by KaptsjaGenerator.py -->       
            <head>
            <meta charset="utf-8" />
            <style>
            {html_header_css}
            </style>
            <script type="text/javascript">
            {gen_javascript}
            </script>
            </head>
            <body>
                <form id="myForm" class="form-asd" role="form" method="POST" action="{captcheck}" enctype="multipart/form-data">
                {modal_header}
                {modal_body}
                    <div>
                        <table width="{width}">
                        <tr><td colspan="3">
                         <!-- Parameter ?random= is required to ensure a reload by some browsers of a new generated image 
                              An unique time stamp is used to avoide any collissions -->
                         <img src="{gen_picture}?random={force_reload}" usemap="#workmap" width="{width}" height="{height}"/>
                           <map name="workmap">{mapareas}
                           </map>
                        </td></tr>
                        <tr><td colspan="3">{number_of_clicks}
                        <input type="text"  id="count_clicks_per_try" name="count_clicks_per_try" size=1 readonly></input>
                        <div  id="message" name="message" ><span></span></div>
                        <input type="hidden" id="count_max_retries" name="count_max_retries" size=1 readonly></input>
                        <input type="hidden" id="reverse" name="reverse" value="{reverse}" size=1 readonly></input>
                        <input type="hidden" id="checkvalue" name="checkvalue" readonly></input>         
                        <input type="hidden" id="mode" name="mode" value="{mode}" readonly></input>         
                        <input type="hidden" id="controlvalue" name="controlvalue" value="{controlvalue}" readonly></input>{exact_nr_circles_submit}
                        </td></tr>
                        <tr><td>
                        {html_reload_button}
                        </td><td>
                        <input type="button" id="retrybutton" name="retrybutton"  value={retry_button_text} onclick="fretry();"></input> 
                        </td><td>
                        <input type="button" form="myForm"  id="submitbutton" name="submitbutton"  value={submit_button_text} onclick="fsubmit();" disabled></input> 
                        </td></tr>
                        <tr><td colspan="3">&nbsp;</td</tr>
                        </table>
                    </div>
                {modal_footer}
                </form>
            </body>
        </html>
        '''.format(width=self.wpic, height=self.hpic, \
        force_reload=str(time.time()), \
        captcheck=captcha_check_url,\
        site_url=site_url, css_url=css_url,  \
        gen_picture=self.generated_unique_picture_file_url, \
        mapareas=''.join(self.map_area_list), \
        controlvalue=self.encodedstring.decode(encoding='UTF-8'), \
        html_header_css=html_header_css, \
        gen_javascript=gen_javascript, \
        html_reload_button=html_reload_button, \
        modal_header=modal_header, \
        modal_body=modal_body, \
        modal_footer=modal_footer, \
        mode=self.mode, reverse=self.reverse,
        number_of_clicks=number_of_clicks,
        exact_nr_circles_submit=exact_nr_circles_submit,
        new_pic_button_text=new_pic_button_text, \
        retry_button_text=retry_button_text, \
        submit_button_text = submit_button_text)
        
    def write_css_file(self):
        # write the css to file, the Kaptsja DIV content HTML file uses this
        try:
            f = open(self.gen_css_file,"w", encoding="utf-8")
            f.write(self.html_header_css())
            f.close()
        except Exception as e:
            logging.error("Writing file %s failed.\n%s" %(self.gen_css_file, e) )
            return 1
        logging.info(r"Created css file for Kaptsja DIV page: %s:" % self.gen_css_file)     
        return 0
                

    def write_javascript_file(self):
        try:
            f = open(self.gen_javascript_file,"w", encoding="utf-8") 
            f.write(self.gen_javascript())
            f.close()
        except Exception as e:
            logging.error("Writing file %s failed.\n%s" %(self.gen_javascript_file, e) )
            return 1
        logging.info(r"Created javascript file for Kaptsja DIV page: %s" % self.gen_javascript_file)
        return 0

    def write_modal_page_file(self):
        soup = bs(self.modal_html(), features=beautifulsoup_parser)  #make BeautifulSoup
        prettyHTML = soup.prettify()   #prettify the html
        try:
            f = open(self.gen_modal_page_file,"w", encoding="utf-8") 
            f.write(prettyHTML)
            f.close()
        except Exception as e:
            logging.error("Writing file %s failed.\n%s" %(self.gen_modal_page_file, e) )
            return 1
        logging.info(r"Created Kaptsja Modal page: %s" % self.gen_modal_page_file)
        return 0

    def write_div_page_file(self):
        soup = bs(self.div_html(), features=beautifulsoup_parser)  #make BeautifulSoup
        prettyHTML = soup.prettify()   #prettify the html
        try:
            f = open(self.gen_div_page_file,"w", encoding="utf-8") 
            f.write(prettyHTML)
            f.close()
        except Exception as e:
            logging.error("Writing file %s failed.\n%s" %(self.gen_div_page_file, e) )
            return 1
        logging.info(r"Created Kaptsja DIV page: %s" % self.gen_div_page_file)
        return 0

    def write_captcha_html_file(self):
        soup = bs(self.captcha_html(), features=beautifulsoup_parser)  #make BeautifulSoup
        prettyHTML = soup.prettify()   #prettify the html
        try:
            f = open(self.gen_captcha_page_file,"w", encoding="utf-8") 
            f.write(prettyHTML)
            f.close()
        except Exception as e:
            logging.error("Writing file %s failed.\n%s" %(self.gen_captcha_page_file, e) )
            return 1
        logging.info(r"Create Kaptsja: generated Kaptsja HTML file: %s" % self.gen_captcha_page_file)
        return 0
        
        logging.info(r"Create Kaptsja: generated Kaptsja HTML file: %s" % self.gen_captcha_page_file)       

    def captcha_thread_remove_files(self):
        for filename in glob.glob(work_dir +'*'):
            generated_filename, generated_file_extension = os.path.splitext(filename)
            generated_filename, unique_time_file = generated_filename.rsplit("_",1)
            try:       
                if float(self.unique_time) > float(unique_time_file) + float(time_keep_captcha_set):
                    try:
                        os.remove(os.path.normpath(filename))
                    except Exception as e:
                        logging.error("Removing file %s failed.\n%s" %(os.path.normpath(filename), e) )
            except:
                pass # probably files with no unique time stamp in directory, those will not be deleted

    def delete_old_files(self):
        t = threading.Thread(target=self.captcha_thread_remove_files())
        t.start()

## functions
def create_captcha(plainstring=None, mode=captcha_page_url):  # defaults to mode for Kaptsja page
    c = Kaptsja(plainstring=plainstring, mode=mode)
    c.create_all_filenames_and_fileurls()
    # The models 2 and 3 work with HTML pages , which must be created first before a Kaptsja will be created and loaded
    if mode == div_page_url:
        c.write_css_file()
        c.write_javascript_file()
        c.write_div_page_file()
        return 0, os.path.normpath(c.gen_div_page_file)
    if mode == modal_page_url:
        c.write_modal_page_file()
        return 0, os.path.normpath(c.gen_modal_page_file)
    c.set_charlist()
    c.create_picturelist()
    c.scale_input_picture()
    c.create_textzone()
    retcode = c.create_circles()
    if retcode > 0:
        return retcode, None
    c.write_captcha_html_file()
    if max_captcha_sets < 1: # when Kaptsja Sets are pre-generated , we do not delete these! 
        c.delete_old_files()
    return 0, os.path.normpath(c.gen_captcha_page_file)

def get_captcha(plainstring=None, mode=captcha_page_url):
    print (mode)
    try:
        captcha_page_list = []
        infile_name = None
        infile_ext = None
        filter = "%s*.html" # only HTML pages to be served!
        if mode == captcha_page_url: 
            infile_name, infile_ext = os.path.splitext(gen_captcha_page)
        elif mode == div_page_url:
            infile_name, infile_ext = os.path.splitext(gen_div_page)
        elif mode == div_captcha_page_url:
            infile_name, infile_ext = os.path.splitext(gen_captcha_page)
        elif mode == modal_page_url:
            infile_name, infile_ext = os.path.splitext(gen_modal_page)
        elif mode == modal_captcha_page_url:
            infile_name, infile_ext = os.path.splitext(gen_captcha_page)
        filter = filter % infile_name     
        for filename in glob.glob(work_dir + filter): 
            captcha_page_list.append(filename)
        if len(captcha_page_list) > 0:
            served_captcha_file = os.path.normpath(random.choice(captcha_page_list))
     # update the time stamps of the pre-generated files, 
     # before serving it as it will help to force reloads, 
     # however when browser caching time of 1 second is not yet exceeded then user must refresh page again by clicking new picture
     # Therefore a short dealy of 1.5 seconds is programmed	search for "setTimeout" in KaptsjaGenerator.py  
     # Try to use this "dirty trick" when there are still caching issues you cannot solve
            #os.utime(os.path.normpath(served_captcha_file))  
            return 0, served_captcha_file
        else:
            logging.error(r"Something went wrong. No Kaptsja sets filtered. Are the Kaptsja sets generated? Run KaptsjaGenerator.py from the command line.")
            return 1, None
    except Exception as e:
            logging.error(r"Something went wrong. Are the Kaptsja sets generated? Run KaptsjaGenerator.py from the command line. Error: %s" % e)
            return 1, None
        

def checkcaptcha(checkvalue, controlvalue, reverse):
    decoded_string = cripto.decryption(controlvalue)
    logging.info(r"Check Kaptsja: checkvalue    : %s" % checkvalue)
    logging.info(r"Check Kaptsja: decoded_string: %s" % decoded_string)
    verify_checkvalue = ""
    inp_sorted = sorted(decoded_string, reverse=int(reverse))
    for i in range(len(inp_sorted)):
        pos = decoded_string.find(str(inp_sorted[i]))
        if pos != -1:
            verify_checkvalue = verify_checkvalue + str(pos)
    logging.info(r"Check Kaptsja: checkvalue       : %s" % checkvalue)
    logging.info(r"Check Kaptsja: verify_checkvalue: %s" % verify_checkvalue)

    if verify_checkvalue == checkvalue:
        return True
    else:
        return False

if __name__ == '__main__':
    if max_captcha_sets > 0:
        print("Please wait till generation of Kaptsja sets has been finished")
        if len(glob.glob(work_dir + "*.html")) > max_captcha_sets:  # there are allready html files generated
            print("Existing Kaptsja sets found. To generate a new set: first delete manually all files in %s" % work_dir)
        if len(glob.glob(work_dir + "*.html")) < 1:  # there are no html files generated when < 1
            for _ in range(max_captcha_sets):
                if active_captcha_model == 1:
                    create_captcha(plainstring=None, mode=captcha_page_url)
                if active_captcha_model == 2:
                    create_captcha(plainstring=None, mode=modal_page_url)
                    create_captcha(plainstring=None, mode=modal_captcha_page_url)
                if active_captcha_model == 3:
                    create_captcha(plainstring=None, mode=div_page_url)
                    create_captcha(plainstring=None, mode=div_captcha_page_url)
            print("Finished!")
        else:
            print("Existing files found (Kaptsja sets?). To generate a new set: first delete manually all files in %s" % work_dir)
            
    else:
        # plainstring = "A123C9" # For test purpose
        plainstring = input("Please provide unique characters from the shown set below.\nAdapt Configuration File when needed. A good number is 6 characters.\nLetters: %s\nDigits : %s\n:" %(used_letters, used_digits))
        # alternative checks
        # if not re.match("^[0-9]*$", plainstring): only digits
        #    logging.info(r"Error! Only digits 0-9 allowed!")
        # if not re.match("^[A-Za-z0-9]+ $", plainstring): #digits and alphabet only      
            # print(r"Error! Only digits 0-9 or alphabet A-Z, a-z are allowed!")
            # sys.exit(syserrmsg)
        for char in plainstring:
            if char not in used_letters + used_digits:
                print ("Character %s is not in the given set. Retry." %char)
                sys.exit(syserrmsg)
            if plainstring.count(char) > 1:
                print ("Character %s occurs more than once which is not allowed. Retry." %char)
                sys.exit(syserrmsg)
        # set the mode you want to test 
        mode=captcha_page_url
        c = Kaptsja(plainstring=plainstring, mode=mode)
        c.create_all_filenames_and_fileurls()
        # The models 2 and 3 work with HTML pages , which must be created first before a Kaptsja will be created and loaded
        if mode == div_page_url:
            c.write_css_file()
            c.write_javascript_file()
            c.write_div_page_file()
            print ("gen_div_page_file", os.path.normpath(c.gen_div_page_file))
        if mode == modal_page_url:
            c.write_modal_page_file()
            print ("gen_modal_page_file", os.path.normpath(c.gen_modal_page_file))
        c.set_charlist()
        c.create_picturelist()
        c.scale_input_picture()
        c.create_textzone()
        retcode = c.create_circles()
        if retcode > 0:
            print ("Return code method create_circles not 0!")
            sys.exit(syserrmsg)
        c.write_captcha_html_file()
        if max_captcha_sets < 1: # when Kaptsja Sets are pre-generated , we do not delete these! 
            c.delete_old_files()
        #
        print("plainstring", c.plainstring)
        print("encodedstring", c.encodedstring)
        print("charlist", c.charlist)
        print("generated_unique_picture_file", c.generated_unique_picture_file)

    ### uncomment this together with the import line for pytesseract at the top to "play" with OCR on the generated pictures
    ### some example are provided. In principle a training data set should be created sot the tool can learn to get a better detection result 
    ### however the generated pictures are very unique, so ti won't be simple
    ###
        # print("*"*10, "Start OCR detection results - eng")
        # print(pytesseract.image_to_string(c.img, lang='eng'))
        # print("-"*10, "next result - eng")
        # print(pytesseract.image_to_string(c.img, lang='eng', config='--psm 7 --oem 3 --dpi 600'))
        # print("-"*10, "next result - eng")
        # print(pytesseract.image_to_string(c.img, lang='eng', config='--psm 6 --oem 3 --dpi 300 -c tessedit_char_whitelist=ABCDEFGHIJKLMNOPQUVWXYZ0123456789'))
        # print("-"*10, "next result - eng")
        # print(pytesseract.image_to_string(c.img, lang='eng', config='--psm 11 --oem 3  -c tessedit_char_whitelist=ABCDEFGHIJKLMNOPQUVWXYZ0123456789') )
        # print("-"*10, "next result - osd")
        # print(pytesseract.image_to_string(c.img, lang='osd', config='--psm 11 --oem 3  -c tessedit_char_whitelist=ABCDEFGHIJKLMNOPQUVWXYZ0123456789') )
        # print("*"*10, "End OCR detection results")



import sys
# sys.path.append("/path/to/your/package_or_module")
sys.path.append(r"./scripts")
from bottle import Bottle, route, run, static_file
from bottle import get, post, request, response, abort, redirect # or route
import KaptsjaGenerator as Captgen
import base64
from bottle import run
import os
import logging
from KaptsjaHTMLpages import *  # we are regenerating the HTML pages so latest settings are used
# import the configuration
from KaptsjaConfiguration import *
#
# uncomment following 2 lines if you have installed gevent and want to use items
# Gevent uses greenlet to provide a high-level synchronous API on top of the libev event loop
# greenlets are lightweight coroutines for in-process sequential concurrent programming
# The main advantage of gevent is its high performance, especially in comparison with traditional threading solutions.
# When gevent is used for network and I/O bound functions which are scheduled cooperatively, it shows its power
# import gevent.monkey;
# gevent.monkey.patch_all()

app = application = Bottle()
# When needed change level level=logging.INFO or level=logging.DEBUG)
# get the directory in which the script was started.
serve_path = os.getcwd()  
try:
    logging.basicConfig(filename=log_file, level=logging.DEBUG, \
        format='%(asctime)s %(levelname)-8s %(message)s', \
        datefmt='%Y-%m-%d %H:%M:%S')
except PermissionError as e:
    print("Check permissions for Kaptsja and its subdirectories.\n\n%s\n\n" % e)
    sys.exit()
except Exception as e:
    print("Start this program from Kaptsja home directory. It was started in %s.\n\n%s\n\n" %(serve_path, e) )
    sys.exit()
##########################################
# START of validation of the configuration
##########################################
syserrmsg = "Error! Look in log file."
if type(siteport) != int:
    text = "siteport %s is not an integer. Change the value." % siteport
    logging.error(text)
    sys.exit(syserrmsg)
if type(sitedebug) != bool:
    text = "sitedebug %s is not an boolean. Change the value to True of False." % sitedebug
    logging.error(text)
    sys.exit(syserrmsg)
if type(site_reloader) != bool:
    text = "site_reloader %s is not an boolean. Change the value to True of False." % site_reloader
    logging.error(text)
    sys.exit(syserrmsg)
##########################################
# END of validation of the configuration
##########################################

# return the home page when url has only a single slash ('/')
@app.route('/')
def index():
    redirect(home_page_url)
 
# return the home page when url equals to site url with and without ending slash  
@app.route(site_url)
@app.route(site_url[:-1]) # without end-slash route also to home page
def index():
    redirect(home_page_url)
    
#############################################################################    
# Start block:  URL requests containing a direct file reference in it (*.html, *.js, *.css, *.png, etcetera) 
# Keep this block before the block URL requests without a file reference.
# return the Kaptsja.ico file when the request url equals to requested favicon.ico file
@app.route('/favicon.ico')
def send_ico_form_media_dir(filename=ico_file):
    return static_file(filename=filename, root=media_dir, mimetype='image/ico')

# return png image file from media_dir when the request url starts with media_url  
@app.route(media_url + '<filename:re:.*\.png>')
def send_image_from_media_dir(filename):
    return static_file(filename, root=media_dir, mimetype='image/png')

# return requested css file from css_dir when the request url starts with ccs_url  
@app.route(css_url + '<filename:re:.*\.css>')
def send_css_from_css_dir(filename):
    return static_file(filename, root=css_dir, mimetype='text/css')

# return javascript file from js_dir when the request url starts with js_url      
@app.route(js_url + '<filename:re:.*\.js>')
def send_javascript_from_js_dir(filename):
    return static_file(filename, root=js_dir, mimetype='text/javascript')

# return requested html file from html_dir when the request url starts with html_url          
@app.route(html_url + '<filename:re:.*\.html>')
def send_html_from_html_dir(filename):
    return static_file(filename, root=html_dir, mimetype='text/html')

# return requested png image file from work_dir when the request url starts with work_url  
@app.route(work_url + '<filename:re:.*\.png>')
def send_image_from_work_dir(filename):
    print("send_image_from_work_dir", filename)
    return static_file(filename, root=work_dir, mimetype='image/png')

# return requested css file from work_dir when the request url starts with work_url      
@app.route(work_url + '<filename:re:.*\.css>')
def send_css_from_work_dir(filename):
    return static_file(filename, root=work_dir, mimetype='text/css')

# return requested javascript file from work_dir when the request url starts with work_url          
@app.route(work_url + '<filename:re:.*\.js>')
def send_javascript_from_work_dir(filename):
    return static_file(filename, root=work_dir, mimetype='text/javascript')

# return requested html file from work_dir when the request url starts with work_url              
@app.route(work_url + '<filename:re:.*\.html>')
def send_html_from_work_dir(filename):
    return static_file(filename, root=work_dir, mimetype='text/html')
#
# END block:  URL requests containing a direct file reference in it

#############################################################################
# Start block: URL requests without a file reference in it.
#
# return the home page when request url starts with home_page_url 
# the home page is created automatically at start of KaptsjaSite.py or when running KaptsjaHTMpages.py
@app.route(home_page_url)
@app.route(site_url + '<filename:path>')
def send_static_home_page(filename=gen_home_page_file):
    return static_file(filename, root='.')

    
# return the success page when request url starts with success_page_url 
# the success page is created automatically at start of ChatChaSite.py or when running KaptsjaHTMpages.py
@app.route(success_page_url)
def send_static_success_page(filename=gen_success_page_file):
    return static_file(filename, root='.')

    
# return the failure page when request url starts with failure_page_url 
# the failure page is created automatically at start of KaptsjaSite.py or when running KaptsjaHTMpages.py
@app.route(failure_page_url)
def send_static_failure_page(filename=gen_failure_page_file):
    return static_file(filename, root='.')


# return the Kaptsja page when request url starts with captcha_page_url 
# the Kaptsja page is created or has been pre-created automatically by KaptsjaGenerator.py
# Captgen.create_captcha : a dynamic page is created for each call
# Captgen.get_captcha    : a pre-generated page is randomly served from the Kaptsja sets 
@app.route(captcha_page_url, method='GET')
@app.route(captcha_page_url, method='POST')
def send_static_captcha_page(filename=gen_captcha_page_file):
    if not active_captcha_model == 1: redirect(home_page_url)
    if max_captcha_sets < 1:
        retcode, gen_captcha_page_file = Captgen.create_captcha(mode=captcha_page_url)
    else:
        retcode, gen_captcha_page_file = Captgen.get_captcha(mode=captcha_page_url)
    response.set_header("Cache-Control", "no-cache, no-store, must-revalidate, max-age=0, s-maxage=0, proxy-revalidate")               
    response.set_header("Pragma", "no-cache")
    response.set_header("Expires", "Sat, 12 Oct 1991 05:00:00 GMT")
    response.set_header("Vary", "*")
    if retcode > 0:
        return "<p>An error has occured. Check the log file.</p>"
    return static_file(filename=gen_captcha_page_file, root='.')


# return the modal page when request url starts with modal_page_url 
# the modal page is created or has been pre-created automatically by KaptsjaGenerator.py 
# Captgen.create_captcha : a dynamic page is created for each call
# Captgen.get_captcha    : a pre-generated page is randomly served from the Kaptsja sets 
@app.route(modal_page_url)
def send_static_modal_page(filename=gen_modal_page_file):
    if not active_captcha_model == 2: redirect(home_page_url)
    if max_captcha_sets < 1:
        retcode, gen_modal_page_file = Captgen.create_captcha(mode=modal_page_url)
    else:
        retcode, gen_modal_page_file = Captgen.get_captcha(mode=modal_page_url)
    print("modal_page_url=", modal_page_url, "gen_modal_page_file=" , gen_modal_page_file) 
    if retcode > 0:
        return "<p>An error has occured. Check the log file.</p>"
    response.set_header("Cache-Control", "no-cache")
    response.set_header("Cache-Control", "must-revalidate, private, max-age=0, no-store")
    return static_file(filename=gen_modal_page_file, root='.')


# return the Kaptsja page (invoked from the modal page) when request url starts with modal_captcha_page_url 
# the Kaptsja page is created or has been pre-created automatically by KaptsjaGenerator.py 
# Captgen.create_captcha : a dynamic page is created for each call
# Captgen.get_captcha    : a pre-generated page is randomly served from the Kaptsja sets 
@app.route(modal_captcha_page_url, method='GET' )
@app.route(modal_captcha_page_url, method='POST' )
def send_static_modal_captcha_page(filename=gen_captcha_page_file):
    if not active_captcha_model == 2: redirect(home_page_url)
    if max_captcha_sets < 1:
        retcode, gen_captcha_page_file = Captgen.create_captcha(mode=modal_captcha_page_url)
    else:
        retcode, gen_captcha_page_file = Captgen.get_captcha(mode=modal_captcha_page_url)
    print("modal_captcha_page_url=", modal_captcha_page_url, "gen_captcha_page_file=" , gen_captcha_page_file) 
    if retcode > 0:
        return "<p>An error has occured. Check the log file.</p>"
    response.set_header("Cache-Control", "no-cache, no-store, must-revalidate, max-age=0, s-maxage=0, proxy-revalidate")
    response.set_header("Pragma", "no-cache")
    response.set_header("Expires", "Sat, 12 Oct 1991 05:00:00 GMT")
    response.set_header("Vary", "*")
    return static_file(filename=gen_captcha_page_file, root='.')


# return the div page when request url starts with div_page_url 
# the div page is created or has been pre-created automatically by KaptsjaGenerator.py 
# Captgen.create_captcha : a dynamic page is created for each call
# Captgen.get_captcha    : a pre-generated page is randomly served from the Kaptsja sets 
@app.route(div_page_url, method='GET')
@app.route(div_page_url, method='POST')
def send_static_div_page(filename=gen_div_page_file):
    if not active_captcha_model == 3: redirect(home_page_url)
    if max_captcha_sets < 1:    
        retcode, gen_div_page_file = Captgen.create_captcha(mode=div_page_url)
    else:
        retcode, gen_div_page_file = Captgen.get_captcha(mode=div_page_url)
    print("div_page_url=",div_page_url, "gen_div_page_file=" , gen_div_page_file) 
    if retcode > 0:
        return "<p>An error has occured. Check the log file.</p>"
    response.set_header("Cache-Control", "no-cache, no-store, must-revalidate, max-age=0, s-maxage=0, proxy-revalidate")               
    response.set_header("Pragma", "no-cache")
    response.set_header("Expires", "Sat, 12 Oct 1991 05:00:00 GMT")
    response.set_header("Vary", "*")
    return static_file(filename=gen_div_page_file, root='.')


# return the Kaptsja page (invoked from the div page) when request url starts with div_captcha_page_url 
# the Kaptsja page is created or has been pre-created automatically by KaptsjaGenerator.py 
# Captgen.create_captcha : a dynamic page is created for each call
# Captgen.get_captcha    : a pre-generated page is randomly served from the Kaptsja sets 
@app.route(div_captcha_page_url, method='GET')
@app.route(div_captcha_page_url, method='POST')
def send_static_div_captcha_page(filename=gen_captcha_page_file):
    if not active_captcha_model == 3: redirect(home_page_url)
    if max_captcha_sets < 1:
        retcode, gen_captcha_page_file = Captgen.create_captcha(mode=div_captcha_page_url)
    else:
        retcode, gen_captcha_page_file = Captgen.get_captcha(mode=div_captcha_page_url)
    if retcode > 0:
        return "<p>An error has occured. Check the log file.</p>"
    response.set_header("Cache-Control", "no-cache, no-store, must-revalidate, max-age=0, s-maxage=0, proxy-revalidate")               
    response.set_header("Pragma", "no-cache")
    response.set_header("Expires", "Sat, 12 Oct 1991 05:00:00 GMT")
    response.set_header("Vary", "*")
    return static_file(filename=gen_captcha_page_file, root='.')


# redirect to the url for the success or failure page when the Kaptsja has been verified resp. true or false
# return to failure in case the checkvalue has length of zero (means not checkvalue was generated, for example due to manual manipulation of the url)
@app.route(captcha_check_url, method='GET')
@app.route(captcha_check_url, method='POST')
def verify_captcha():
    checkvalue = request.forms.get('checkvalue')
    controlvalue = request.forms.get('controlvalue')
    reverse = request.forms.get('reverse')
    mode = request.forms.get('mode')
    if len(checkvalue) != number_of_letters + number_of_digits:
        redirect(failure_page_url)
    else:
        retcode = Captgen.checkcaptcha(checkvalue, controlvalue, reverse)
        if retcode == True:
            redirect(success_page_url)
        else:
            redirect(failure_page_url)
    return static_file(staticFile, filePath)
 
 
# here we redirect to the home page, normally that would be your next logical site page 
# this url is set by the Success page "Click here to continue"  
@app.route(success_cont_url)
def send_success_cont_url():
    redirect(site_url)


#  here we redirect to the home page, normally that would be a retry or no access page 
# this url is set by the Failure page "Click here to continue"     
@app.route(failure_cont_url)
def send_failure_cont_url():
    redirect(site_url)
#    
# END block: URL requests without a file reference in it.
#
if __name__ == "__main__":
    try:        
        retcode = create_KaptsjaHome_html_file()
        if retcode == True:
           print(r"Generation of Kaptsja Home HTML page file succeeded.")
        else:
           print(r"Generation of Kaptsja Home HTML page file failed.")

        retcode = create_KaptsjaSuccess_html_file()
        if retcode == True:
           print(r"Generation of Kaptsja Success HTML page file succeeded.")
        else:
           print(r"Generation of Kaptsja Success HTML page file failed.")

        retcode = create_KaptsjaFailure_html_file()
        if retcode == True:
           print(r"Generation of Kaptsja Failure HTML page file succeeded.")
        else:
           print(r"Generation of Kaptsja Failure HTML page file failed.")
        
        run(app, server='%s' % siteserver, host='%s' % sitehost, port=siteport, debug=sitedebug, reloader=site_reloader)
    except Exception as e:
        text = "\nRunning the bottle Web server failed. Is there a Python error message? Check that first then all start up setttings. Is sitehost correct (name or IP address)?\nPython error message is:\n%s" % e
        logging.error(text)
        print (text)



###################################################################################################
#  KaptsjaEncDec uses AES Encryption to encrypt / decrypr ANY data string or file 
#  Kaptsja is the Dutch phonetic pronunciation of the English word Captcha.
#  In Kaptsja it encrypts/decrypts the controlvalue send with the Kaptsja page.
#  This encryption/ decryption module can be used universally in many projects!
#
#  WARNING: BEFORE CHANGING ANYTHING IN THIS FILE MAKE A BACKUP SO YOU CAN RESTORE !
#  Read the comments in this file as they contain valuable information to understand the effects 
###################################################################################################

from Crypto.Cipher import AES
from Crypto import Random
import base64
import hashlib
import os
import random

key_filename = "secret_key.txt"

def get_create_save_key(self):
    if os.path.exists(key_filename) == True:
        kfile = open(key_filename,"rb")
        key = kfile.readline()
        kfile.close()
        # logging.info(r'Get key: Encryption key: %s' % secret_key)
    else:
        kfile = open(key_filename,"wb") 
        # os.urandom generates a random value of 32 bytes long using key_size
        # the key size may be 128, 192, or 256 bits long.
        # 256 bits equals to 32 bytes 
        # 128 bits equals to 16 bytes 
        key_size = 32
        key = os.urandom(key_size)
        kfile.write(key)
        kfile.close() 
        # logging.info(r'Create & Save Key: Encryption key: %s' % key)
    return key


class Cripto(object): #  a variant of Crypto :-)
    def __init__(self, key_filename): 
        self.key_filename = key_filename
        # get the key from the key file
        if os.path.exists(self.key_filename) == True:
            kfile = open(self.key_filename,"rb")
            self.key = kfile.readline()
            kfile.close()
            # logging.info(r'Get key: Encryption key: %s' % secret_key)
        else:
            kfile = open(self.key_filename,"wb") 
            # os.urandom generates a random value of 32 bytes long using key_size
            # the key size may be 128, 192, or 256 bits long.
            # 256 bits equals to 32 bytes 
            # 128 bits equals to 16 bytes 
            key_size = 32
            self.key = os.urandom(key_size)
            kfile.write(self.key)
            kfile.close() 
            # logging.info(r'Create & Save Key: Encryption key: %s' % key)
        #A block cipher works on units of a fixed size (known as a block size), but messages come in a variety of lengths. 
         # block size MUST be equal to 16 which is the value of AES.block_size. 
        # Do not change this!
        self.block_size = AES.block_size
        # MODE_CBC: Ciphertext Block Chaining is a mode of operation where each
        # plaintext block gets XOR-ed with the previous ciphertext block prior to encryption.
        # XOR-ed is an exclusive or, or exclusive disjunction and is a logical operation 
        # that outputs true only when inputs differ (one is true, the other is false).
        self.mode_cbc = AES.MODE_CBC
        
    #################################### ENCRYPTION METHODS ##################################### 
    
    def encryption(self, data):  # all encryption steps in one go
        ''' 
        The correct sequence of processing for encryption, 
        For easy following the flow, all intermediate data results are numbered and deleted at the end. 
        This is not per se needed and all data1,data2,.. may be called just data)
        Input data is a default Python 3 string of type str which is by default encoded with utf-8
        Steps for encription:
         A. Encode the still readable data (of Python type str) into bytes; a "bytes literal"), 
            this will ensure correct length calculation when double-byte or multiple bytes characters (sets) have been used like Chinese characters.
         B. Add padding characters to data (of Python type bytes)) till a length of a multiple of 16 bytes has been reached.
         C. Encrypt with cipher the encoded data to a encrypted bytes data object using default encoding set 'utf-8'
         D. Concatenate the initiation vector and the encrypted bytes data object together to a single data value in bytes.
            This concatenated data (iv_and_enc) is the bytes (of Python type bytes) representation of the encrypted end result. 
            The encryption output can be written into a binary file (write/read file with open(name, mode="wb") or open(name, "wb")
         G. Encode with library Base64 the concatenated initiation vector and encrypted bytes object to a "readable" string 
            This Base64 encode output is a string (of Python type str) representation of the encrypted end result as is printable and transferrable (e.g. via http).
        '''
        data1 = self.encode_string(data)                # encode the still readable input string of of Python 3 type str which defaults to utf-8.
        data2 = self.pad_bytes(data1)                   # pad the encoded bytes  - when needed - with extra bytes till a multiple of 16 bytes has been reached 
        data3 = self.encrypt_string(data2)              # encrypt the encoded bytes literal (a multiple of 16 bytes)  
        self.iv_and_enc = self.iv + data3               # concatenate initiation vector (bytes) + encrypted bytes literal to a bytes literal (= the bytes Encrypted end result)
        data4 = self.base64encode_string(self.iv_and_enc) # create with based64 encode a printable/transferrable string version of it in Python 3 type str which defaults to utf-8.
        encryption_result = data4                       # return encryption result as string
        del data, data1, data2, data3, data4            # delete all intermediate data to quickly save memory 
        return encryption_result   
        
    def encode_string(self, data):  # in: an normal string of Python 3 type str which defaults to utf-8. may contain double/multiple bytes characters like 一個長句子，外加一些額外的單詞。
        # encode printable characters to bytes
        return data.encode()  #out: encoded bytes literal of Python 3 type bytes
 
    def pad_bytes(self, data): # in: should always be an encoded byte string (not to be confused with base64 encoded!!)
        extra = len(data) % self.block_size
        if extra > 0:
            data = data + b"\0" * (self.block_size - extra)
        return data # out: a padded encoded byte string of a multiple of 16 bytes length ready for encryption!

    def encrypt_string(self, data):  # in: encoded padded bytes literal
        # if following error occurs it means that the input is NOT properly padded: ValueError: Input strings must be a multiple of 16 in length
        #  Usually this happens when data of type(data) == str has been padded BEFORE encoding, which is wrong
        #  Always pad and unpad on bytes not on strings! 
        #   -  There is only one exception and that is when all characters in the original input string are single byte characters!  
        #   -  Then padding of original input string after the encode would also result in a multiple of 16 bytes length. Avoid this method!   
        #   
        # create initiation vector for encryption in Python 3 format bytes
        self.iv = Random.new().read(self.block_size)
        # create cipher for encryption and decription
        self.cipher = AES.new(self.key, self.mode_cbc, self.iv)
        encrypted_string =  self.cipher.encrypt(data) # use cipher to encrypt the bytes
        return encrypted_string   # out: encrypted bytes literal 

    def base64encode_string(self, iv_and_enc):  # in: concatenated iv and encrypted bytes literal (Note: the input here is NOT only the output of self.function encrypt_string()!
        # encode concatenated iv and the encrypted bytes literal to a printable/transferrable string of Python 3 type str
        # The iv and the encryted data in bytes MUST be kept together else decryption will always fail (therefore concatenation is the preferred method) 
        return base64.b64encode(iv_and_enc) # out: base64 encoded string literal of Python 3 type str
    
    # SOME HELPER FUNCTIONS

    def get_iv(self):  # returns no value (None) when not set
        if self.iv: return self.iv
        
    def get_cipher(self): # returns no value (None) when not set
        if self.cipher: return self.cipher



    #################################### DECRYPTION METHODS ##################################### 
    
    def decryption(self, data): # all decryption steps in one go
        ''' 
        The correct sequence of processing for decryption, 
        For easy following the flow, all intermediate data results are numbered and deleted at the end. 
        This is not per se needed and all data1,data2,.. may be called just data)
        Input is a printable and transferrable (e.g. via http) representation of the encrypted text string (equals to the output of encryption process as defined above).
        Steps for decription:
         A. Decode with library Base64 the concatenated initiation vector and encrypted string to bytes
         B. Decrypt with cipher the Base64 decoded result to a bytes object (still padded)
         C. Remove padding characters to decypted bytes.
         D. Decode the unpadded bytes into a readable string (original text)
        '''
        data1 = self.base64decode_string(data)    # input is a printable/transferrable base64 encoded str, which must be decoded to bytes first
        data2 = self.decrypt_string(data1)           # decrypt decoded bytes literal to decrypted bytes
        data3 = self.unpad_bytes(data2)              # remove all padding characters from decrypted bytes         
        data4 = self.decode_string(data3)            # decode the unpadded bytes to original string
        decryption_result = data4                    # return encryption result
        del data, data1, data2, data3, data4              # delete all intermediate strings to quickly save memory 
        return decryption_result

    def base64decode_string(self, data): #in: base64encoded string of Python 3 type str. This contains the concatenated iv and encrypted data
        # decode the printable/tranferrable characters from string Python 3 type str into Python 3 type bytes
        return base64.b64decode(data)  #out: decoded bytes literal

    def decrypt_string(self, data): # in: base64decoded bytes literal, which when originally correctly created is already a multiple of 16 bytes length
        iv = data[:self.block_size] # first 16 bytes of the bytes literal contains the initiation vector, which together with the key is needed for decryption
        cipher = AES.new(self.key, self.mode_cbc, iv) # determine the cipher based on key and iv and encryption mode
        enc = data[self.block_size:]  # After the first 16 bytes of the initiation vector, the encryted data follows
        encoded_string = cipher.decrypt(enc)   # decrypt the encryted data
        return encoded_string    # return the decripted but still padded data as bytes
    
    def unpad_bytes(self, data):                # in: decrypted but still padded data as bytes
        unpadded_string =  data.rstrip(b"\0")   # get rid of all padding characters (which were added before)
        return unpadded_string                  # out decrypted unpadded data as bytes

    def decode_string(self, data):   #in: encoded and unpadded data in bytes
        # decode printable ASCII characters to bytes
        return data.decode('utf-8')  #out: decoded string of data in Python type str, which should be equal to the original input for encryption!


    #################################### HELPER AND FILE METHODS ##################################### 
        
    def str2bytes(self, data):        # in: a  string
        return bytes(data, 'utf-8')   # out : a bytes literal
        
    def bytes2str(self, data):        # in : a bytes literal
        return str(data)     # out : a  string

    def encrypt_file(self, file_name):
        with open(file_name, 'r', encoding="utf8") as fo:
            text_from_file = fo.read()
        encrypted_text = self.encryption(text_from_file)
        with open(file_name[:-4] + "_enc.txt", 'wb') as fo:
            fo.write(encrypted_text)

    def decrypt_file(self, file_name):
        with open(file_name[:-4] + "_enc.txt", 'rb') as fo:
            encrypted_text_from_file = fo.read()
        decrypted_text = self.decryption(encrypted_text_from_file)
        with open(file_name[:-4] + "_dec.txt", 'w', encoding="utf-8") as fo:
            fo.write(decrypted_text)

if __name__ == '__main__':
    import sys

###################################################################
#    Start Encryption and Decryption in ONE GO             
###################################################################
    cript = Cripto(key_filename)
    text ="A long sentence with some extra words. 一個長句子，外加一些額外的單詞。"
    # direct one go encryption
    encrypted_result = cript.encryption(text)
    print ("\nEncrypted result\n", encrypted_result , "\n")

    # direct one go decryption
    decrypted_result = cript.decryption(encrypted_result)
    print ("\nDecrypted result\n", decrypted_result, "\n")

###################################################################
#    Encryption and Decryption in a Step by Step approach             
###################################################################
#    ENCRYPTION - Step by Step           
###################################################################

    print ("*"*20, "ENCRYPTION STARTED step by step", "*"*20)    
  
    text ="A 一個長句子 long NEW sentence with some extra words. 一個長句子，外加一些額外的單詞。"
    print("\nInput: original text (str)\n", text, "\n")
    
    # Show initiation vector and corresponding chiper
    print("\nOutput: inition vector (bytes)\n", cript.get_iv() )
    print("Output: chiper (bytes)", cript.get_cipher(), "\n")
    
    text1 = cript.encode_string(text)          # encode the text from string to bytes
    print("\nOutput: encoded text (bytes)\n", text1, "\n")

    text2 = cript.pad_bytes(text1)              # pad the encoded bytes text 
    print("\nOutput: padded text (bytes)\n", text2, "\n")
       
    text3 = cript.encrypt_string(text2)         # encrypt encoded bytes literal
    print("\nOutput: encrypted text (bytes), excluded initiaton vector!\n", text3, "\n")

    iv_and_enc = cript.iv + text3               # concatenate initiation vector + encrypted bytes literal
    print("\nOutput: encryption result (bytes) equals to initiaton vector+encrypted text \n", iv_and_enc, "\n")

    text4 = cript.base64encode_string(iv_and_enc) # create a printable/transferrable version of it
    encryption_output = text4                  # keep encryption_output for decryption process

    print("\nOutput: encryption result (str in base64 format) equals to initiaton vector+encrypted text\n", encryption_output, "\n")
    del text1, text2, text3, text4             # delete all intermediate strings saving memory 

###################################################################
#    DECRYPTION - Step by Step          
###################################################################

    print ("*"*20, "DECRYPTION STARTED  step by step", "*"*20)    

    text5 = encryption_output
    print("\nInput: encryption result (str in base64 format) equals to initiaton vector+encrypted text\n", text5, "\n")

    text6 = cript.base64decode_string(text5)   # input is string: printable/transferrable  
    print("\nOutput: base64 decoded text (bytes)\n", text6, "\n")

    text7 = cript.decrypt_string(text6)       # decrypt encoded bytes to padded bytes
    print("\nOutput: decrypted text (bytes)\n", text7, "\n")

    text8 = cript.unpad_bytes(text7)         # delete padding characters to get original text
    print("\nOutput: unpadded text (bytes)\n", text8, "\n")
 
    text9 = cript.decode_string(text8)        # decode padded bytes to padded string    
    decryption_result = text9      
           # return encryption result
    del text5, text6, text7, text8             # delete all intermediate strings saving memory 
    print("\nOutput: decoded text = decryption result, should be equal as original text (str)\n", decryption_result, "\n") 


###################################################################
#    Encryption and Decryption of a File Content in ONE GO             
###################################################################

    # Create a file with some text to be encrypted 
    # Prefix "Z__" is used to find file easyly back when sorting.
    # Files created with prefix "Z__" can be deleted after reviewing test results
    
    cript = Cripto(key_filename)

# -----------------------------------------------------------------
#    Create the input file in text mode  (mode="w")
#    The input file  cannot be a binary file!
#    It must be created with: open(<filename>, 'w', encoding="utf8")  <! mandatory for text with certain language characters    
#                         or: open(<filename>, 'w')                   <! 
#   The default encoding is platform dependent (whatever import locale; locale.getpreferredencoding() returns
# -----------------------------------------------------------------    
    test_filename = 'Z__input.txt'
    text ="A long NEW FILE sentence with some extra words. 一個長句子，外加一些額外的單詞。"
    with open(test_filename, 'w', encoding="utf8") as fo:
        fo.write(text)
              
    cript.encrypt_file( test_filename)
    cript.decrypt_file( test_filename) 
    print ("\nTest: input text file and corresponding encryption and decryption result files are created in the run directory.\n")



import setuptools

with open("Kaptsja/README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Kaptsja", # Replace with your own username
    version="10.1.4",
    author="Peter Stamps",
    author_email="p.j.m.stamps@gmail.com",
    description="Kaptsja is the Dutch 'phonetic pronunciation' for this CaptCha software with special and complex features.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/peterstamps/Kaptsja/",
    project_urls={
    'Documentation': 'https://github.com/peterstamps/Kaptsja'},
    # packages=setuptools.find_packages(),
    packages=setuptools.find_packages(include=['Kaptsja', 'Kaptsja.*']),
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Operating System :: OS Independent",
    ],
    keywords='Captcha catpcha, recaptcha, re-captcha Encryption Decryption Picture Puzzle Circle Character Generator',
    tag='http bottle web server proxy http proxy server proxy server http server http web server proxy framework web framework Python3 catpcha recaptcha re-captcha hcaptcha google cloud mitm man in the middle web server web scraping botting bottle encrypt encryption decrypt decryption AES picture image PIL pillow puzzle',
    license_file  = "Kaptsja Copyright Notice.txt",
    install_requires=[],
    python_requires='>=3.7.4',
    scripts=['Kaptsja/scripts/KaptsjaSite.py', 'Kaptsja/scripts/KaptsjaGenerator.py', 'Kaptsja/scripts/KaptsjaConfiguration.py', 'Kaptsja/scripts/KaptsjaHTMLpages.py', 'Kaptsja/scripts/KaptsjaPictureIco.py', 'Kaptsja/scripts/KaptsjaEncDec.py'],
    package_dir={'': 'Kaptsja/scripts'},
    data_files=[
                ('Kaptsja/.', ['Kaptsja/Kaptsja Copyright Notice.txt', 'Kaptsja/Readme.rst','Kaptsja/Start_Kaptsja_website.bat','Kaptsja/Start_Kaptsja_website.sh']),
                ('Kaptsja/css', ['Kaptsja/css/bootstrap.min-3.3.7.css']),
                ('Kaptsja/docs', ['Kaptsja/docs/Installation_of_Kaptsja_with_ginx_and_uwsgi.rst', 'Kaptsja/docs/Readme2.rst']),
                ('Kaptsja/html', ['Kaptsja/html/KaptsjaFailurePage.html', 'Kaptsja/html/KaptsjaHome.html', 'Kaptsja/html/KaptsjaSuccessPage.html']),
                ('Kaptsja/js', ['Kaptsja/js/bootstrap.min-3.3.7.js', 'Kaptsja/js/jquery.min-3.5.1.js']),
                ('Kaptsja/key', ['Kaptsja/key/place_holder_delete_me.txt']),
                ('Kaptsja/log', ['Kaptsja/log/Kaptsja.log']),
                ('Kaptsja/media', ['Kaptsja/media/Glass_1.jpg', 'Kaptsja/media/Glass_2.jpg', 'Kaptsja/media/Glass_3.jpg', 'Kaptsja/media/Glass_4.jpg', 'Kaptsja/media/Glass_5.jpg', 'Kaptsja/media/Glass_6.jpg', 'Kaptsja/media/Glass_7.jpg', 'Kaptsja/media/Glass_8.jpg', 'Kaptsja/media/Kaptsja.ico', 'Kaptsja/media/Kaptsja_bg.jpg']),
                ('Kaptsja/media/randomlist', ['Kaptsja/media/randomlist/Glass_1.jpg', 'Kaptsja/media/randomlist/Glass_2.jpg', 'Kaptsja/media/randomlist/Glass_3.jpg', 'Kaptsja/media/randomlist/Glass_4.jpg', 'Kaptsja/media/randomlist/Glass_5.jpg', 'Kaptsja/media/randomlist/Glass_6.jpg', 'Kaptsja/media/randomlist/Glass_7.jpg', 'Kaptsja/media/randomlist/Glass_8.jpg', ]),
                ('Kaptsja/scripts', ['Kaptsja/scripts/KaptsjaSite.py', 'Kaptsja/scripts/KaptsjaGenerator.py', 'Kaptsja/scripts/KaptsjaConfiguration.py', 'Kaptsja/scripts/KaptsjaHTMLpages.py', 'Kaptsja/scripts/KaptsjaPictureIco.py', 'Kaptsja/scripts/KaptsjaEncDec.py','Kaptsja/scripts/Z__input.txt', 'Kaptsja/scripts/Z__input_dec.txt', 'Kaptsja/scripts/Z__input_enc.txt']),
                ('Kaptsja/work', ['Kaptsja/work/place_holder_delete_me.txt']),
                ],
)


###################################################################################################
#  WARNING: BEFORE CHANGING ANYTHING IN THIS FILE MAKE A BACKUP SO YOU CAN RESTORE !
#  Read the comments in this file as they contain valuable information to understand the effects 
###################################################################################################
import sys                                             # keep this import!
#
# There are 3 models from which you can select: Note: on home page URLs for other models are set to href="#" (deactivated) 
# Set captcha_model to 1, 2 or 3!
#  1. default    -> the Kaptsja is shown direct on a single full HTML page
#  2. modal      -> the Kaptsja is "injected" in a modal window generated by bootstrap.js at top of a full HTML page
#  3. div        -> the Kaptsja is "injected" in a single full HTML page using JQuery.js 
#
active_captcha_model        = 3           # Default: 2. Only 1, 2 or 3 allowed. It defaults to 2 when not set or when > 3 or < 1 
#                                         # When changing to another model make work_dir empty and 
#                                         # Run Kaptsjagenerator.py from command line to generate the corresponding Kaptsja sets
#                                         # See also max_captcha_sets   
max_captcha_sets            = 0           # Recommended 20 or higher: the maximum number of pre-generated Kaptsja sets (html,png,css,js files)
#                                         # When max_captcha_sets = 0 only dynamic Kaptsja generation happens!
#                                         # When the number of Kaptsja sets is low there will be a higher chance that
#                                         # the same Kaptsja will be shown twice or more in a sequence!
time_keep_captcha_set       = 60          # Value in seconds. Each Kaptsja set is uniquely generated in the work_dir to avoid 
#                                         # concurrency collisions. It will be kept there for the seconds defined.
#                                         # When max_captcha_sets > 0 then the time_keep_captcha_set value is not used and
#                                         # all generated files in work_dir will be kept
#
input_picture               = r"Kaptsja_bg.jpg" # Always keep a copy of your input picture as it might be rescaled due to width size or  setting of input_picture_rescale_width 
input_picture_rescale_width = 0                 # Default 0. Zero means no rescaling else use value between 360 - 598 
#                                               # When set input_picture will be resized to this pixel size and overwritten!!
#                                               # Be carefull as 360 is really a low number. Fonts below 30 for circles and 10 for text would easily needed
#                                               # Font default locations (Windows 10 and Ubuntu 20.04 are used here)
if(sys.platform=='win32'):
    font_circle             = r"c:/windows/fonts/Arial.ttf"   # Window fonts location
#    font_circle             = r"c:/windows/fonts/MSJHBD.ttc"   # Window fonts location for Chinese Traditional Han Microsoft JhengHei Bold (has been tested as well)
    font_textzone           = r"c:/windows/fonts/Arial.ttf"        
else:
    font_circle             = r"/usr/share/fonts/truetype/msttcorefonts/Arial.ttf"   # Linux: To locate Arial font use command: fc-list | grep arial 
    font_textzone           = r"/usr/share/fonts/truetype/msttcorefonts/Arial.ttf"     
font_pnts_circle            = 45              # Font points are not only about the size (width and height!)
font_pnts_textzone          = 20              # About Font points:
#                                             # In typography, the point is the smallest unit of measure. 
#                                             # It is used for measuring font size, leading, and other items on a printed page. 
#                                             # The size of the point has varied throughout the history of printing.
used_letters    = u"ABCDEFGHJKLMNPRSTUVWXYZ"  # Randomly given number of letters will be shown on Kaptsja. 
#                                             # Use u" " for Special Unicode Characters like u"個長句子。". NOTE: to display
#                                             # special characters like Chinese the correct (Chinese) font must be specified!
number_of_letters           = 2               # Default: 2 A maximum of 10 characters is allowed 
#                                             # (total of number_of_letters + number_of_digits < 11
used_digits                 = r"0123456789"   # Numbers 6 and 9 will be underscored when rotating is active
number_of_digits            = 4               # Default 4: Recommended: number_of_letters + number_of_digits is 6
#                                             # The number of circles drawn equals to number_of_letters + number_of_digits
#                                             # An even number 2 is better for precise positioning of area circle on HTML page 
max_retries                 = 2               # Counted after first try!, the maximum number people may retry to solve the captcha
border_circle               = 2               # The border thickness of circles in pixels
border_line                 = 3               # The thickness for lines in pixels
line_color                  = "#808080"       # String: default #808080 gray #d93211 orange/red HTML color names with Hex color codes
rotate_char_circle          = True            # Default: True. Set to False for better performance
rotate_angle_char_circle    = 75              # Default: 30. Means between -30 and +30  Max value is 360.
circle_color_trans          = 15              # Value between 0 and 100. No % sign! Higher number means more transparency. 50 is half.
char_circle_color_trans     = 0               # Value between 0 and 100. No % sign! As character color equals  the inversed circle color,
#                                             # a higher number here has the opposite effect!! 
#                                             # 100 means the full inversed circle color will be shown on the character: usefull for higher contrast
#                                             # 0 means full transparency. This is usefull is the background picture is relative dark.
#                                             # You get a see through the circle effect!
textzone_color_trans        = 50              # Value between 0 and 100. No % sign! Higher number means more transparency. 50 is half.
textzone_center_text        = True            # True or False (Python bool) True=center, False=Left Aligned
textzone_text_color         = "#fafafa"       # String:  HTML color names with Hex color codes
textzone_bg_color           = "#040406"       # String:  HTML color names with Hex color codes
max_morphing_pie_slices     = 9               # Minimal 3, Default 15 which is sufficient for Arial fontpoints 35
#                                             # A higher number of max_morphing_pie_slices will morph characters too much and they might become unreadable 
copyright_text_picture      =  "© Margrhet"   # Copyright text put at upper right or left corner of picture
copyright_text_color        = "#f123ab"       # String:  HTML color names with Hex color codes
copyright_text_position     = "R"             # R/r=right or L/l=Left upper corner of picture. Default R
font_pnts_copyright_text    = 12              # Using same font as for textzone

#                                                         
# Please check if the sub-directories do exist under the home_dir Kaptsja, only the key_dir will be automatically created if it doesn't exist
# Directory settings
home_dir          = r"./"                                 # current directory-> end slash required (in this directory KaptsjaSite.py should be started)
css_dir           = home_dir + r"css/"                    # directory path for javascript files -> end slash required
docs_dir          = home_dir + r"docs/"                   # directory path for documentationfiles (txt, doc, docx, rst, ...) -> end slash required
html_dir          = home_dir + r"html/"                   # directory path for HTML files -> end slash required
js_dir            = home_dir + r"js/"                     # directory path for javascript files -> end slash required
media_dir         = home_dir + r"media/"                  # directory path for media files (jpg, png, ...) -> end slash required
randomlist_dir    = media_dir + r"randomlist/"            # directory path for media files (jpg, png, ...) to show randomly.
#                                                         # -> end slash required. 
#                                                         # This works only when randomlist is set to True 
randomlist        = True                                  # Kaptsja Pictures from randomlist will be used (randomly selected)
scripts_dir       = home_dir + r"scripts/"                # directory path for python scripts  -> end slash required
key_dir           = home_dir + r"key/"                    # directory path for encryption key -> end slash required
log_dir           = home_dir + r"log/"                    # directory path for log file -> end slash required
work_dir          = home_dir + r"work/"                   # directory path for encryption key -> end slash required
# 
gen_picture       = r"KaptsjaPicture.png"       # Name of generated Kaptsja picture 
gen_captcha_page  = r"KaptsjaPage.html"         # Name of generated Kaptsja HTML page 
gen_home_page     = r"KaptsjaHome.html"         # Name of generated Kaptsja Home page  (home page of the site)
gen_div_page      = r"KaptsjaDIV.html"          # Name of generated Kaptsjadiv page 
gen_modal_page    = r"KaptsjaModal.html"        # Name of generated Kaptsja Modal page 
gen_success_page  = r"KaptsjaSuccessPage.html"  # Name of generated Kaptsja Success page  
gen_failure_page  = r"KaptsjaFailurePage.html"  # Name of generated Kaptsja Failure page 
gen_javascript    = r"KaptsjaDIV.js"           # Name of generated javascript; script of generated Kaptsja DIV page  
gen_css           = r"KaptsjaDIV.css"          # Name of generated css; style of generated Kaptsja page  
#
bootstrap_js_file    = js_dir + "bootstrap.min-3.3.7.js"    # Name / location of the used bootstrap javascript file
bootstrap_css_file   = css_dir + "bootstrap.min-3.3.7.css" # Name / location of the used bootstrap css file
jquery_js_file       = js_dir + "jquery.min-3.5.1.js"       # Name / location of the used jquery javascript file

# Full Path File names
secret_key_file          = key_dir   + r"Kaptsja_secret_key.txt"  # Path to the encryption / decryption key for controlvalue
log_file                 = log_dir   + r"Kaptsja.log"        # Path to log file 
ico_file                 = media_dir + r"Kaptsja.ico"        # Path to log file 
input_picture_file       = media_dir + input_picture # Path to input picture 
# Although generated (once generated when KaptsjaSite.py or KaptsjaHTMLpages.py are run) 
gen_home_page_file       = html_dir  + gen_home_page         # Path to generated Kaptsja Home page of the site
gen_success_page_file    = html_dir  + gen_success_page      # Path generated Kaptsja Success page 
gen_failure_page_file    = html_dir  + gen_failure_page      # Path generated Kaptsja Failure page 
# These files are generated once (Kaptsja Sets; delete manual from work_dir) or multiple times (dynamic; deleted automatically)
gen_picture_file         = work_dir  + gen_picture           # Path to generated Kaptsja picture
gen_captcha_page_file    = work_dir  + gen_captcha_page      # Path to generated Kaptsja HTML page
gen_div_page_file        = work_dir  + gen_div_page          # Path to generated Kaptsja DIV page 
gen_modal_page_file      = work_dir  + gen_modal_page        # Path to generated Kaptsja Modal page 
gen_javascript_file      = work_dir  + gen_javascript        # Path to generated javascript; script of generated Kaptsja page
gen_css_file             = work_dir  + gen_css               # Path generated css; style of generated Kaptsja page  
#
# Settings for Bottle example program KaptsjaSite.py
sitehost = "localhost"                         # hostname or IP address must be between quotes
siteport = 8080                                # port number must be integer, number below 1000 might require special Admin rights
siteserver = "wsgiref"                         # Values: Default "wsgiref" , use "python_server" when in production with nginx and uwsgi, or "gevent"
sitedebug = True                               # Values: True, False. Bottle shows debugging messages in start cmd window when True. Us False in production.
site_reloader = True                           # Values: True, False. When changes in loaded Python code are detected Bottle will automatically reload it.
#
# URL to resources or invoking module functions
# URL parts settings
site_url                    = r"/captsite/"                 # base URL for Kaptsja site    -> end slash is required
css_url                     = site_url + r"css/"            # URL part for all css files  -> end slash required
docs_url                    = site_url + r"docs/"           # URL part for all documentation files (txt, doc, docx, rst, ...) -> end slash required
html_url                    = site_url + r"html/"           # URL part for all HTML files  -> end slash required
js_url                      = site_url + r"js/"             # URL part for all javascript files (url part) -> end slash required
media_url                   = site_url + r"media/"          # URL part for all media files (png,jpg, ...) -> end slash required
work_url                    = site_url + r"work/"           # URL part for all work files (png,jpg, ...) -> end slash required
home_page_url               = site_url + "home"             # URL part for HREF reference/routing to Home Page
captcha_page_url            = site_url + "captcha"          # URL part for HREF reference/routing to generated Kaptsja HTML page
div_page_url                = site_url + "div"              # URL part for HREF reference/routing to generated Kaptsjadiv page  
div_captcha_page_url        = site_url + "div-Kaptsja"      # URL part for HREF reference/routing to generated Kaptsja Modal Page
modal_page_url              = site_url + "modal"            # URL part for HREF reference/routing to generated Kaptsja Modal Page
modal_captcha_page_url      = site_url + "modal-captcha"    # URL part for HREF reference/routing to refresh Kaptsja in genereated Kaptsja Modal Page
success_page_url            = site_url + "success"          # URL part for HREF reference/routing to Kaptsja Success Page
failure_page_url            = site_url + "failure"          # URL part for HREF reference/routing to Kaptsja Failure Page
captcha_check_url           = site_url + "check-captcha"    # URL part for HREF reference/routing to captcha_check_function
success_cont_url            = site_url + "success-continue" # URL part for HREF reference/routing to continue after display Kaptsja Success Page
failure_cont_url            = site_url + "failure-continue" # URL part for HREF reference/routing to continue after display Kaptsja Failure Page
gen_javascript_url          = js_url + gen_javascript       # URL to generated javascript ; contains script of generated Kaptsja page
gen_css_url                 = css_url + gen_css             # URL to generated javascript ; contains script of generated Kaptsja page
#
beautifulsoup_parser        = "html.parser"                 # Default html.parser. When lxml has been installed then you can specify "lxml"
#                                                           # Parser  used to prettify and 'complete' generated html pages
# ###################################################################
# TEXT SECTION The following text is displayed on the Kaptsja Pages. You May Change them into your language.
####################################################################
# When text literal is set to None, no text will be shown
# When text literal is set to "", the default English text will be shown
# Else the "translated text" set here will be shown
#
check_the_box_text =  "Please check box to continue"
number_of_clicks = "Number of clicks:"
too_many_clicks_text = "Too many clicks. Retry."
exact_nr_circles_submit =  "Clicking the exact number of circles enables Submit button"
submitted_text = "Submitted. Click button New Picture."
too_many_retries_text = "Too many retries. Click button New Picture."
new_pic_button_text = "New Picture"
retry_button_text = "Retry"
submit_button_text = "Submit Result"
intro_captcha_text = "CaptCha for Human recognition"
captcha_hdr_text = "Human Verification"
# Close modal window character X or use another HTML entity defined as &xxx; 
# Symbol &times; equals to &#215;  text may be used style is H2
captcha_hdr_close = "&otimes;"     
not_a_robot_text= "I am a not a robot&nbsp;&nbsp;"
## Button colors
submit_button_bgcolor_enabled = "#a3ff9e"   # greenish
submit_button_bgcolor_disabled = "#ff9e9e"  # redish
retry_button_bgcolor_enabled = "#a3ff9e"   # greenish
retry_button_bgcolor_disabled = "#ff9e9e"  # redish
new_pic_button_bgcolor_enabled = "#a3ff9e"   # greenish
new_pic_button_bgcolor_disabled = "#ff9e9e"  # redish
# Home page text
home_page_tab1 = "Home"
home_page_tab2 = "1. Kaptsja Page"
home_page_tab3 = "2. Kaptsja Modal Page"
home_page_tab4 = "3. Kaptsja inside page element &lt;div&gt ... &lt;/div&gt;"
home_page_click_tab_text = "To select the activated Kaptsja model click on tab "
home_page_one_model_active_text = "Only one model can be active at a time. To activate another model, "
home_page_change_setting_text = "change setting 'active_captcha_model' in configuration file KaptsjaConfiguration.py."

# Following instruction text will be displayed. Do not change the Sorting Order Instruction itself else failures will happen!
# These are the sorting instructions, You may not change the order or Digits and Letters as listed below!
# When you put \n inside the text a hard next line break will happen and the textzone will be adapted automatically
# as well as the positioning of the circles.
# Always keep this in mind:
# in ASCENDING order (▲). First Digits then Letters
# in DESCENDING order (▼). First Letters then Digits.
text_Ascending = "Are you human? Click in each circle. First Digits then Letters.\n In ASCENDING order (▲).\n Not readable? Click on New Picture button."
text_Descending = "Are you human? Click in each circle. First Letters then Digits.\n In DESCENDING order (▼).\n Not readable? Click on New Picture button."
# ###################################################################
# Performance setting. Read text first before changing this setting!
allowed_evals_per_char    = 250     # Default 250 --> increase only step-by-step! 
# CAUTION: Adapt this value when not all circles are drawn on a given picture in combination with the given font sizes. 
# Reason: A small height of the picture leads to higher number of evaluations to find non-overlapping space.
# Increasing this number might help a little, but leads to performance decrease!
# An increase of picture height is a better solution or reduce the font size, which is less optimal for users. 
# Backgound information: 
# For each circle a random position is generated and verified that there is no overlap with other circles. 
# To avoid endless loops or too much evaluations a maximum number is calculated as follows: 
# Maximum evaluations = "Allowed evaluations per character" multiplied by sum of "letters and digits" as defined above. 
# When the maximum is reached a break happens in the loop and an error message is generated in the log file.
# The error message contains a suggested minimum height for your picture! 
# Always test first when using changing pictures, font type or size and total number of characters.
# ###################################################################



###################################################################################################
#  KaptsjaHTMLpages creates the KaptsjaHome page, KaptsjaSuccess page and KaptsjaFilure page
#  Kaptsja is the Dutch phonetic pronunciation of the English word Captcha.
#
#  WARNING: BEFORE CHANGING ANYTHING IN THIS FILE MAKE A BACKUP SO YOU CAN RESTORE !
#  Read the comments in this file as they contain valuable information to understand the effects 
###################################################################################################

import sys
# sys.path.append("/path/to/your/package_or_module")
sys.path.append(r"./scripts")
import os
import time
import logging
# import the configuration
from KaptsjaConfiguration import *
############################################################################################################################################################
#  SETTINGS  (start) See /Kaptsja/scripts/KaptsjaConfiguration.py 
############################################################################################################################################################
# When needed change level level=logging.INFO or level=logging.DEBUG)
# get the directory in which the script was started.
serve_path = os.getcwd()  
try:
    logging.basicConfig(filename=log_file, level=logging.ERROR)
except PermissionError as e:
    print("Check permissions for Kaptsja and its subdirectories.\n\n%s\n\n" % e)
    sys.exit()
except Exception as e:
    print("Start this program from Kaptsja home directory. It was started in %s.\n\n%s\n\n" %(serve_path, e) )
    sys.exit()
############################################################################################################################################################
#  SETTINGS  (End)
############################################################################################################################################################

##########################################
# START of validation of the configuration
##########################################  

if 360 <= int(input_picture_rescale_width) <= 598:
    modal_content_width = input_picture_rescale_width + 2 
if int(input_picture_rescale_width) == 0: # 
    modal_content_width = 600  # this is the default value
text = "Home"
if not "home_page_tab1" in globals():
    home_page_tab1 = text
if home_page_tab1 == "":
    home_page_tab1 = text
if home_page_tab1 is None:
    home_page_tab1 = ""

text = "1. Kaptsja Page"
if not "home_page_tab2" in globals():
    home_page_tab2 = text
if home_page_tab2 == "":
    home_page_tab2 = text
if home_page_tab2 is None:
    home_page_tab2 = ""
    
text = "2. Kaptsja Modal Page"
if not "home_page_tab3" in globals():
    home_page_tab3 = text
if home_page_tab3 == "":
    home_page_tab3 = text
if home_page_tab3 is None:
    home_page_tab3 = ""

text = "3. Kaptsja inside page element &lt;div&gt ... &lt;/div&gt;"
if not "home_page_tab4" in globals():
    home_page_tab4 = text
if home_page_tab4 == "":
    home_page_tab4 = text
if home_page_tab4 is None:
    home_page_tab4 = ""

text = "To select the activated Kaptsja model click on tab "
if not "home_page_click_tab_text" in globals():
    home_page_click_tab_text = text
if home_page_click_tab_text == "":
    home_page_click_tab_text = text
if home_page_click_tab_text is None:
    home_page_click_tab_text = ""

text = "Only one model can be active at a time. To activate another model, "
if not "home_page_one_model_active_text" in globals():
    home_page_one_model_active_text = text
if home_page_one_model_active_text == "":
    home_page_one_model_active_text = text
if home_page_one_model_active_text is None:
    home_page_one_model_active_text = ""

text = "change setting 'active_captcha_model' in configuration file KaptsjaConfiguration.py."
if not "home_page_change_setting_text" in globals():
    home_page_change_setting_text = text
if home_page_change_setting_text == "":
    home_page_change_setting_text = text
if home_page_change_setting_text is None:
    home_page_change_setting_text = ""
    
##########################################
# END of validation of the configuration
##########################################

def create_KaptsjaHome_html_file():
    # All braces for css and javascript itself MUST be double notation like {{ // some css or javascript code }} to escape the first. 
    # Only ony brace will be generated in the output. All formatting keys like .format(key= ) MUST have a single brace {key} notation.
    captcha_page_url_active = "#"
    modal_page_url_active = "#"
    div_page_url_active = "#"
    if active_captcha_model == 1:
        captcha_page_url_active = captcha_page_url 
    elif active_captcha_model == 2:
        modal_page_url_active = modal_page_url 
    elif active_captcha_model == 3:
        div_page_url_active = div_page_url 
    
    html_1 = '''
    <html>
<!-- this file is generated by KaptsjaHTMLpages.py -->
    <head>
    <!-- Include Bootstrap 3.3.7 CSS here in the head section -->
    <link rel="stylesheet" href="/captsite/css/bootstrap.min-3.3.7.css"> 
    <style>
    .nav-tabs > li > a {{
    margin-right: 2px;
    line-height: 1.42857143;
    border-left: 1px solid #00000038;
    border-right: 1px solid #00000038;
    border-top: 1px solid #00000038;
    border-radius: 4px 4px 0 0;
    }}
    </style>
    </head>
    <body>
    <div>
        <ul class="nav nav-tabs">
          <li role="presentation" class="active"><a href="#">{home_page_tab1}</a></li>
          <li role="presentation"><a href="{captcha_page_url}">{home_page_tab2}</a></li>
          <li role="presentation"><a href="{modal_page_url}">{home_page_tab3}</a></li>
          <li role="presentation"><a href="{div_page_url}">{home_page_tab4}</a></li>
        </ul>
    </div>
    <div><h3>{home_page_click_tab_text} {active_captcha_model}.</h3>
    <h5>{home_page_one_model_active_text}</h5>
    <h5>change setting 'active_captcha_model' in configuration file KaptsjaConfiguration.py.<h5></div>
    <!-- Include First the jQuery and Second Bootstrap javascript libraries. Order is important! -->
    <script src="/captsite/js/jquery.min-3.5.1.js"></script>
    <script src="/captsite/js/bootstrap.min-3.3.7.js"></script>
     </body>
</html>     
    '''.format(captcha_page_url=captcha_page_url_active, \
    modal_page_url=modal_page_url_active, \
    div_page_url=div_page_url_active, \
    active_captcha_model=active_captcha_model, \
    home_page_click_tab_text=home_page_click_tab_text, \
    home_page_one_model_active_text=home_page_one_model_active_text, \
    home_page_change_setting_text=home_page_change_setting_text, \
    home_page_tab1=home_page_tab1, \
    home_page_tab2=home_page_tab2, \
    home_page_tab3=home_page_tab3, \
    home_page_tab4=home_page_tab4
    )

    try:
        htmlfile = open(gen_home_page_file,"w", encoding="utf-8") 
        htmlfile.write(html_1)
        htmlfile.close()
        logging.info(r"Generation Kaptsja Home HTML page file success: %s" % gen_home_page_file)
        return True
    except:
        logging.error(r"Generation Kaptsja Home HTML page file failure: %s" % gen_home_page_file)
        return False

def create_KaptsjaSuccess_html_file():
    # All braces for css and javascript itself MUST be double notation like {{ // some css or javascript code }} to escape the first. 
    # Only ony brace will be generated in the output. All formatting keys like .format(key= ) MUST have a single brace {key} notation.
    html='''<!doctype html>
<html>
<!-- this file is generated by KaptsjaHTMLpages.py -->
  <head>
  </head>
    <style>
    body {{
        text-align: center;
        padding: 40px 0;
        background: #EBF0F5;
    }}
    h1 {{
        color: #88B04B;
        font-family: "Nunito Sans", "Helvetica Neue", sans-serif;
        font-weight: 900;
        font-size: 40px;
        margin-bottom: 10px;
    }}
    p {{
        color: #404F5E;
        font-family: "Nunito Sans", "Helvetica Neue", sans-serif;
        font-size:20px;
        margin: 0;
    }}
    .card {{
        background: white;
        padding: 60px;
        border-radius: 4px;
        box-shadow: 0 2px 3px #C8D0D8;
        display: inline-block;
        margin: 0 auto;
    }}
    .checkmark {{
        font-family: "Nunito Sans", "Helvetica Neue", sans-serif;
        color: #ffffff;
        font-size: 150px;
        line-height: 200px;
        margin-left: 0px;;
    }}
    </style>
    <body>
      <div class="card">
      <div style="border-radius:200px; height:200px; width:200px; background: #9ABC66; margin:0 auto;">
        <span class="checkmark">✓</span>
      </div>
        <h1>Success</h1> 
        <p>You solved the CaptCha successfully;<br/> You may now continue. Click <a href="{success_cont_url}">here.</a></p>
      </div>
    </body>
</html>
    '''.format(success_cont_url=success_cont_url)
    
    try:
        htmlfile = open(gen_success_page_file,"w", encoding="utf-8") 
        htmlfile.write(html)
        htmlfile.close()
        logging.info(r"Generation Kaptsja Success HTML page file success: %s" % gen_success_page_file)
        return True
    except:
        logging.error(r"Generation Kaptsja Success HTML page file failure: %s" % gen_success_page_file)
    return False

def create_KaptsjaFailure_html_file():
    # All braces for css and javascript itself MUST be double notation like {{ // some css or javascript code }} to escape the first. 
    # Only ony brace will be generated in the output. All formatting keys like .format(key= ) MUST have a single brace {key} notation.
    html='''<!doctype html>
<html>
<!-- this file is generated by KaptsjaHTMLpages.py -->
  <head>
  </head>
    <style>
    body {{
        text-align: center;
        padding: 40px 0;
        background: #EBF0F5;
    }}
    h1 {{
        color: #f54242;
        font-family: "Nunito Sans", "Helvetica Neue", sans-serif;
        font-weight: 900;
        font-size: 40px;
        margin-bottom: 10px;
    }}
    p {{
        color: #404F5E;
        font-family: "Nunito Sans", "Helvetica Neue", sans-serif;
        font-size:20px;
        margin: 0;
    }}
    .card {{
        background: white;
        padding: 60px;
        border-radius: 4px;
        box-shadow: 0 2px 3px #C8D0D8;
        display: inline-block;
        margin: 0 auto;
    }}
    .falsemark {{
        font-family: "Nunito Sans", "Helvetica Neue", sans-serif;
        color: #ffffff;
        font-size: 150px;
        line-height: 200px;
        margin-left: 0px;;
    }}
    </style>
    <body>
      <div class="card">
      <div style="border-radius:200px; height:200px; width:200px; background: #f54242; margin:0 auto;">
        <span class="falsemark">X</span>
      </div>
        <h1>Failure</h1> 
        <p>You did not solve the CaptCha successfully;<br/> You may now retry. Click <a href="{failure_cont_url}">here.</a></p>
      </div>
    </body>
</html>
    '''.format(failure_cont_url=failure_cont_url)
    
    try:
        htmlfile = open(gen_failure_page_file,"w", encoding="utf-8") 
        htmlfile.write(html)
        htmlfile.close()
        logging.info(r"Generation Kaptsja Failure HTML page file success: %s" % gen_failure_page_file)
        return True
    except:
        logging.error(r"Generation Kaptsja Failure HTML page file failure: %s" % gen_failure_page_file)
        return False

if __name__ == '__main__':
    retcode = create_KaptsjaHome_html_file()
    if retcode == True:
       print(r"Generation of Kaptsja Home HTML page file succeeded.")
    else:
       print(r"Generation of Kaptsja Home HTML page file failed.")

    retcode = create_KaptsjaSuccess_html_file()
    if retcode == True:
       print(r"Generation of Kaptsja Success HTML page file succeeded.")
    else:
       print(r"Generation of Kaptsja Success HTML page file failed.")

    retcode = create_KaptsjaFailure_html_file()
    if retcode == True:
       print(r"Generation of Kaptsja Failure HTML page file succeeded.")
    else:
       print(r"Generation of Kaptsja Failure HTML page file failed.")
